﻿using System;
using System.Runtime.InteropServices;
using System.Text;

/*
 * BasicSpectrometer (C# Windows)
 * -----------------
 * Copyright 2023 by Avenir Photonics GmbH & Co. KG
 * 
 * This is a very simple demo to show the essential steps for measuring a spectrum.
 * The spectrometer is controlled via USB.
 * 
 * Besides the main program, there are two more classes:
 * 
 *  - WinUsbInterface is a wrapper for WinUSB, the generic USB driver provided by Microsoft
 *    (this class was reduced to the minimum features required for this demo)
 *    
 *  - ZioLinkProtocol is a very simple implementaion of the ZioLink protocol.
 *    It uses WinUsbInterface to communicate with the device.
 *  
 *  This demo is intentionally kept as simple as possible in order to show the basic concepts.
 *  Any real application should probably include more error checking and a timeout.
 * 
 */

namespace BasicSpectrometerConsole
{
    internal class Program
    {
        static void Main(string[] args)
        {
            Console.WriteLine("BasicSpectrometer");
            Console.WriteLine("-----------------");

            try
            {
                //******************** Open USB connection to spectrometer ********************

                Console.WriteLine("Searching for spectrometer:");

                ZioLinkProtocol ziolink = null;
                string[] devpaths = WinUsbInterface.FindInterfaces();
                foreach (string devpath in devpaths)
                    if (devpath.StartsWith("\\\\?\\usb#vid_354f&pid_01")) //if Vendor ID is 0x354F and Product ID is 0x01xx
                    {
                        ziolink = new ZioLinkProtocol(devpath);
                        Console.WriteLine("Found at device path: " + devpath);
                        break; //if several spectrometers ar present: only use the first one
                    }
                if (ziolink == null)
                {
                    Console.WriteLine("No device found.");
                    return;
                }

                ziolink.Open(); //Open USB interface

                ziolink.SendReceiveMessage(0x0000); //0x0000 = Send Command: Reset

                //******************** Read device information ********************

                byte[] modelName = ziolink.SendReceiveMessage(0x2003); //0x2003 = Get Device Property: ModelName
                Console.WriteLine("Model name: " + Encoding.ASCII.GetString(modelName).TrimEnd('\0'));

                byte[] serialNo = ziolink.SendReceiveMessage(0x2001); //0x2001 = Get Device Property: SerialNumber
                Console.WriteLine("Serial number: " + Encoding.ASCII.GetString(serialNo).TrimEnd('\0'));

                int pixelCount = BitConverter.ToInt32(ziolink.SendReceiveMessage(0x2007)); //0x2007 = Get Device Property: PixelCount
                Console.WriteLine("Number of pixels: " + pixelCount);

                //******************** Read wavelengths ********************

                float[] c = new float[4]; //4 wavelength coefficients
                for (int i = 0; i < c.Length; i++)
                    c[i] = BitConverter.ToSingle(ziolink.SendReceiveMessage(0x201C + i)); //0x201C = Get Device Property: WavelengthCoeff0
                float[] wavelengths = new float[pixelCount];
                for (int p = 0; p < pixelCount; p++)
                    wavelengths[p] = c[0] + (c[1] + (c[2] + c[3] * p) * p) * p;
                Console.WriteLine("Wavelength range: " + wavelengths[0].ToString("0.00") + " to " + wavelengths[pixelCount - 1].ToString("0.00") + " nm");

                //******************** Set exposure parameters ********************

                Console.WriteLine("Setting exposure parameters to: 20 ms, 10x averaging, no auto exposure");
                ziolink.SendReceiveMessage(0x1100, 20000); //0x1100 = Set Parameter: ExposureTime (in microseconds)
                ziolink.SendReceiveMessage(0x1101, 10); //0x1101 = Set Parameter: Averaging
                ziolink.SendReceiveMessage(0x1109, 0); //0x1109 = Set Parameter: AutoExposureEnabled

                //******************** Start Exposure ********************

                Console.WriteLine("Starting exposure ...");
                ziolink.SendReceiveMessage(0x0004, 1); //0x0004 = Send command: StartExposure (with parameter number of spectra = 1)

                //******************** Wait for exposure to be finished ********************

                SpectrStatus status;
                int availableSpectra;
                do
                {
                    int st = BitConverter.ToInt32(ziolink.SendReceiveMessage(0x3000)); //0x3000 = Get Measured Value: Status
                    status = (SpectrStatus)(st & 0xFF); //byte 0 of returned value
                    availableSpectra = (st >> 8) & 0xFFFF; //byte 1 and 2 of returned value
                } while (status == SpectrStatus.TakingSpectrum);
                Console.WriteLine("Exposure finished. " + availableSpectra + " spectrum available in device.");

                //******************** Read spectrum from device ********************

                Console.WriteLine("Reading spectrum from device:");
                byte[] rawdata = ziolink.SendReceiveMessage(0x4000); //0x4000 = Get Bulk Data: Spectrum
                if (rawdata.Length != 64 + pixelCount * 4) throw new Exception("Unexpected number of bytes in received spectrum."); //should never happen

                //The first 64 bytes are the spectrum header (see struct definition below)
                GCHandle handle = GCHandle.Alloc(rawdata, GCHandleType.Pinned);
                SpectrumHeader header = (SpectrumHeader)Marshal.PtrToStructure(handle.AddrOfPinnedObject(), typeof(SpectrumHeader));
                handle.Free();

                float[] spectrum = new float[header.PixelCount];
                Buffer.BlockCopy(rawdata, 64, spectrum, 0, pixelCount * 4); //Copy byte buffer to integer array (offset 64 = size of SpectrumHeader)

                //******************** Display values ********************

                //Now we can read all values transmitted in the spectrum header, for example:
                Console.WriteLine("Load level: " + header.LoadLevel.ToString("0.00") + " %");

                Console.WriteLine("First 10 spectrum values:");
                for (int p = 0; p < 10; p++)
                    Console.WriteLine(wavelengths[p].ToString("0.0") + " nm: " + spectrum[p]);

                Console.WriteLine("Done.");
            }
            catch (Exception ex)
            {
                Console.WriteLine("An error occurrd: " + ex.Message);
            }
        }

        //***************** Structures and Enumerations ********************

        [StructLayout(LayoutKind.Sequential, Pack = 1, CharSet = CharSet.Ansi)]
        private struct SpectrumHeader
        {
            public uint ExposureTime; //in us (0: unknown)
            public uint Averaging; //(0: unknown)
            public uint Time; //elapsed 0.1 ms since midnight (0xFFFFFFFF = unknown)
            public uint Date; //elapsed days (0xFFFFFFFF = unknown)
            public float LoadLevel; //1.0: Saturation Level (< 0: unknown or n/a)
            public float SensorTemperature; //in °C (< 274: unknown)
            public ushort PixelCount;
            public byte PixelFormat; //0x00 (reserved for future use)
            public byte VersionNumber; //0x00 (reserved for future use)
            public ushort ExposureSettings; 
            public ushort AppliedProcessing; 
            [MarshalAs(UnmanagedType.ByValTStr, SizeConst = 24)]
            public string Name; //up to 3 null-terminated strings: Spectrum name, intensity unit, sample name
            public float DarkAvg;
            public float ReadoutNoise;
        }

        public enum SpectrStatus
        {
            Idle = 0x00,
            WaitingForTrigger = 0x01,
            TakingSpectrum = 0x02,
            WaitingForTemperature = 0x03,
            PoweredOff = 0x08,
            SleepMode = 0x09,
            NotConnected = 0x0A
        }
    }
}
