﻿using System;

namespace DeviceClasses
{
    /// <summary>This abstract base class includes basic methods and properties shared by all communication interfaces.</summary>
    /// <remarks>The basic communication scheme is half-duplex: The host sends a message to the device and then the
    /// device replies with a message to the host. Afterwards, the host can send a new message. There is no out-of-band communication.
    /// All interfaces include some way to determine the length of a message.</remarks>
    public abstract class CommunicationInterface
    {
        /// <summary>Opens the interface to a specific device.</summary>
        /// <remarks>The device is specified by the <see cref="DevicePath"/> property.</remarks>
        public abstract void Open();

        /// <summary>Closes the interface to a specific device.</summary>
        /// <remarks>The device is specified by the <see cref="DevicePath"/> property.</remarks>
        public abstract void Close();

        /// <summary>Indicates whether the interface is open.</summary>
        /// <value><c>true</c> if the interface is open, otherwise <c>false</c>.</value>
        public abstract bool IsOpen { get; }

        /// <summary>A string that uniquely indentifies the communication interface and the device address, if applicable.</summary>
        /// <remarks>This property is set by the constructor.</remarks>
        public abstract string DevicePath { get; }

        /// <summary>Writes data to the interface.</summary>
        /// <param name="buffer">A byte buffer that contains the data to be sent (starting at index 0).</param>
        /// <param name="count">The number of bytes to be sent.</param>
        /// <remarks>The <see cref="Buffer"/> is supplied by the caller and intended to be reused.</remarks>
        public abstract void Write(byte[] buffer, int count);

        /// <summary>Reads data from the interface.</summary>
        /// <param name="buffer">A byte buffer to store the incoming data.</param>
        /// <param name="count">The maximum number of bytes to be read.</param>
        /// <returns>The number of bytes that are actually read.</returns>
        /// <remarks>The size of the <see cref="Buffer"/> determined the maximum number of bytes to be read.</remarks>
        /// <remarks>The <see cref="Buffer"/> is supplied by the caller and intended to be reused.</remarks>
        public abstract int Read(byte[] buffer, int count);

        //potential extensions: Timeout, Purge
    }

    public interface IUsbInterface
    {
        /// <summary>The name of the USB device.</summary>
        /// <remarks>According to the Device Descriptor.</remarks>
        string Product { get; }

        /// <summary>The manufacturer of the USB device.</summary>
        /// <remarks>According to the Device Descriptor.</remarks>
        string Manufacturer { get; }

        /// <summary>The serial number of the USB device.</summary>
        /// <remarks>According to the Device Descriptor.</remarks>
        string SerialNumber { get; }
    }

    public interface ITextInterface
    {
        string LineTerminator { get; set; }

        string ReadLine();

        void WriteLine(string text);
    }

    [Serializable]
    public class CommunicationException : Exception
    {
        public CommunicationException() { }
        public CommunicationException(string message) : base(message) { }
        public CommunicationException(string message, Exception innerException) : base(message, innerException)  { }
    }
}