﻿using System;

namespace DeviceClasses
{
    /// <summary>This abstract base class contains basic methods and properties shared by all devices.</summary>
    /// <remarks>All implementations of this interface should include a static method <c>FindDevices()</c> that takes no arguments and returns
    /// an array of device class instances. This method should search and return any device that can be automatically found without further
    /// configuration parameters.</remarks>
    public abstract class Device
    {
        /// <summary>Opens the interface to this device.</summary>
        /// <remarks>The device is specified by the <see cref="DevicePath"/> property, which is set by the constructor.</remarks>
        public abstract void Open();

        /// <summary>Closes the interface to this device.</summary>
        /// <remarks>The device is specified by the <see cref="DevicePath"/> property.</remarks>
        /// <remarks>This method does not throw any exceptions, for example if the devices is already closed or was removed.</remarks>
        public abstract void Close();

        /// <summary>Indicates whether the interface is open.</summary>
        /// <value><c>true</c> if the interface is open, otherwise <c>false</c>.</value>
        public abstract bool IsOpen { get; }

        /// <summary>A string that uniquely indentifies the communication interface and the device address, if applicable.</summary>
        /// <remarks>This property is set by the constructor. It corresponds to the <see cref="ICommunicationInterface.DevicePath"/> property.</remarks>
        public abstract string DevicePath { get; }

        /// <summary>The name of this device.</summary>
        public abstract string DeviceName { get; }

        /// <summary>The manufacturer of this device.</summary>
        public abstract string Manufacturer { get; }

        /// <summary>The serial number of this device.</summary>
        public abstract string SerialNumber { get; }

        /// <summary>A name that uniquely indentifies this device.</summary>
        /// <remarks>This is usually the <see cref="DeviceName"/>, followed by the <see cref="SerialNumber"/> and/or the port name.</remarks>
        /// <remarks>This name can be used to check if two devices classes refer to the same physical device.</remarks>
        public override string ToString() { return DeviceName; }

        public static bool operator ==(Device obj1, Device obj2)
        {
            if (ReferenceEquals(obj1, obj2)) return true;
            if (ReferenceEquals(obj1, null)) return false;
            return obj1.Equals(obj2);
        }

        public static bool operator !=(Device obj1,Device obj2)
        {
            if (ReferenceEquals(obj1, obj2)) return false;
            if (ReferenceEquals(obj1, null)) return true;
            return !obj1.Equals(obj2);
        }

        public override bool Equals(object obj)
        {
            Device other = obj as Device;
            if (ReferenceEquals(obj, null)) return false;
            if (ReferenceEquals(this, obj)) return true;
            return other.ToString() == this.ToString();
        }
    }

    /// <summary>An interface for thermometers.</summary>
    /// <remarks>This is an interface and not an abstract base class because many devices may include a temperature sensor in addition to their primary function.</remarks>
    public interface IThermometer
    {
        /// <summary>Gets the current temperature.</summary>
        /// <value>The actual temperature in °C.</value>
        /// <remarks>A temperature of less than -274 °C indicates a sensor fault.</remarks>
        /// <remarks>This property is called "actual" temperature, because if the temperature is controlled, there is also a target temperature.</remarks>
        float ActualTemperature { get; }
    }

    /// <summary>An interface for devices controlled by the ZioLink protocol.</summary>
    public interface IZioLinkProtocol
    {
        string ReadString(int commandCode);
        int ReadInt(int commandCode);
        int SendReceiveMessage(byte[] sendMessge, byte[] replyBuffer); //returns number of bytes received (including return code)
        void Close();
    }

    public abstract class Spectrometer : Device, IThermometer
    {
        public abstract int PixelCount { get; }

        public abstract float[] Wavelengths { get; }

        public abstract float ExposureTime { get; set; }
        public abstract float MinExposureTime { get; }
        public abstract float MaxExposureTime { get; }

        public abstract int Averaging { get; set; }
        public abstract int MaxAveraging { get; }

        public abstract float ActualTemperature { get; }

        /// <summary>Sends a request to take spectra to the spectrometer.</summary>
        /// <param name="NumSpectra">The number of spectra to be taken. Usually this is just 1.
        /// Special cases: -1 means take spectra continuously, but keep only the most recent spectrum.
        /// -2 also takes spectra continously, but keeps all spectra as long as there is space in the FIFO buffer.
        /// </param>
        public abstract void StartExposure(int NumSpectra);
        public virtual void StartExposure() { StartExposure(1); }

        public abstract void CancelExposure();

        /// <summary>Get the Status and the AvailableSpectra.</summary>
        /// <param name="AvailableSpectra">The number of spectra that are available now.</param>
        /// <returns>The current status of the spectrometer.</returns>
        public abstract SpectrStatus GetStatus(out int AvailableSpectra);
        public virtual SpectrStatus GetStatus() { return GetStatus(out _); }

        public abstract SpectrumData GetSpectrumData();
        public abstract float[] GetSpectrum();
    }

    public interface IAutoExposure
    {
        bool AutoExposure { get; set; }
        float AutoExposureTime { get; set; }
    }

    public interface ICalibrationSpectra
    {
        bool SubtractDark { get; set; }
        SpectrIntensityUnits IntensityUnit { get; set; }
        void TakeDarkSpectrum();
        void TakeReferenceSpectrum();
    }

    public enum SpectrStatus
    {
        Idle = 0x00,
        WaitingForTrigger = 0x01,
        TakingSpectrum = 0x02,
        WaitingForTemperature = 0x03,
        PoweredOff = 0x08,
        SleepMode = 0x09,
        NotConnected = 0x0A,
    }

    public enum SpectrIntensityUnits
    {
        Counts = 0x0,
        CountsPerMs = 0x1,
        Percentage = 0x2,
        Absorbance = 0x3,
        Calibrated = 0x4
    }

    public class SpectrumData : Object
    {
        public float[] Values = null;
        public float ExposureTime = 0;
        public int Averaging = 1;
        public DateTime TimeStamp = DateTime.Now; //refers to start of exposure
        public float LoadLevel = 0;
        public float SensorTemperature = -300;      
        public int ExposureSettings = 0;
        public ProcessingValues AppliedProcessing = 0;
        public string Name = "";
        public string IntensityUnit = "";
        public string Sample = "";
        public float DarkAvg = 0;
        public float ReadoutNoise = 0;

        public int PixelCount { get { return Values.Length; } }

        [Flags]
        public enum ProcessingValues
        {
            IntensityUnit_Counts = 0x0,
            IntensityUnit_CountsPerMs = 0x1,
            IntensityUnit_Percentage = 0x2,
            IntensityUnit_Absorbance = 0x3,
            IntensityUnit_Calibrated = 0x4,
            IntensityUnit_Evaluated = 0x5,
            DarkSpectrum_None = 0x00,
            DarkSpectrum_Subtracted = 0x10,
            DarkSpectrum_SubtractedTempComp = 0x30,
            IntensityUnit_Mask = 0xF,
            DarkSpectrum_Mask = 0xF0
        }
    }
}