﻿using System;
using System.Collections.Generic;
using System.Diagnostics;
using System.Runtime.InteropServices;
using System.Text;

namespace DeviceClasses
{
    public class ZioLinkSpectrometer : Spectrometer, IZioLinkProtocol, IAutoExposure, ICalibrationSpectra
    {
        private static Guid DeviceInterfaceGuid = new Guid("{34EC51E0-5052-462B-A5A7-5DC0AF784A71}");
        private DateTime date0 = new DateTime(1970, 1, 1);

        private Version minFirmwareDateTime = new Version(4, 0, 7, 1);
        private Version minFirmwareCalibrSpectra = new Version(4, 0, 8, 2);

        private enum DeviceType { Unknown = -1, Default = 0 }
        private int[] devPID = new int[] { 0x0200 };

        //*************** Initialization and Configuration ****************

        public static ZioLinkSpectrometer[] FindDevices()
        {
            List<ZioLinkSpectrometer> devices = new List<ZioLinkSpectrometer>();

            string[] devpaths = WinUsbInterface.FindInterfaces();
            foreach (string devpath in devpaths)
                if (devpath.StartsWith("\\\\?\\usb#vid_354f&pid_01")) //if product id is 0x01xx
                    if (!devpath.StartsWith("\\\\?\\usb#vid_354f&pid_0103")) //but not 0x0103 (NioLink)
                    {
                        ZioLinkSpectrometer dev = new ZioLinkSpectrometer(devpath);
                        if (dev.IsAvailable) devices.Add(dev);
                    }

            return devices.ToArray();
        }

        public ZioLinkSpectrometer(string devicePath)
        {
            usb = new WinUsbInterface(devicePath);
            IsAvailable = usb.IsAvailable;
        }

        protected WinUsbInterface usb;
        public bool IsAvailable { get; }

        public override void Open()
        {
            if (IsOpen) Close();
            usb.Open();
            try
            {
                Write((int)Cmd.Reset);
                softwareVersion = ReadVersion((int)Cmd.GetSoftwareVersion);

                pixelCount = ReadInt((int)Cmd.GetPixelCount);

                int n = 4; //so far only 4 wavelength coefficients supported
                float[] c = new float[n]; // { 0, 1, 0, 0 };
                for (int i = 0; i < n; i++) c[i] = ReadFloat((int)Cmd2.GetWavelengthCoeff0 + i);
                wavelengths = new float[PixelCount];
                for (int p = 0; p < pixelCount; p++) wavelengths[p] = c[0] + (c[1] + (c[2] + c[3] * p) * p) * p;

                minExposureTime = ReadInt((int)Cmd.GetMinExposureTime) * 1E-6f;
                maxExposureTime = (float)Math.Floor(ReadInt((int)Cmd.GetMaxExposureTime) * 1E-6f); //round down to seconds (prevent invalid exposure times due to rounding)
                exposureTime = ReadInt((int)Cmd.GetExposureTime) * 1E-6f;
                maxAveraging = ReadInt((int)Cmd.GetMaxAveraging);
                averaging = ReadInt((int)Cmd.GetAveraging);
                autoExposure = (ReadInt((int)Cmd2.GetAutoExposureEnable) != 0);
                autoExposureTime = ReadInt((int)Cmd2.GetAutoExposureTime) * 1E-6f;
                subtractDark = (ReadInt((int)Cmd2.GetSubtractDark) != 0);
                intensityUnit = (SpectrIntensityUnits)ReadInt((int)Cmd2.GetIntensityUnit);

                UpdateTimeInDevice();

                float voltage = ReadFloat((int)Cmd2.GetSupplyVoltage);
                if (voltage < 4.3) throw new Exception("Supply voltage is too low (" + voltage.ToString("0.00") + " V)."); //We allow another 0.1 V because of measurement uncertainty
                if (voltage > 5.6) throw new Exception("Supply voltage is too high (" + voltage.ToString("0.00") + " V)."); //We allow another 0.1 V because of measurement uncertainty
            }
            catch
            {
                usb.Close();
                throw;
            }
        }

        public override void Close()
        {
            if (usb.IsOpen)
            {
                try
                {
                    Write((int)Cmd.StandBy);
                }
                catch (Exception ex)
                {
                    System.Diagnostics.Debug.WriteLine(ex.Message + " while closing device.");
                }
                usb.Close();
            }
        }

        //*************** Device Properties ****************

        public override bool IsOpen { get { return usb.IsOpen; } }

        public override string DevicePath { get { return usb.DevicePath; } }

        public override string DeviceName { get { return usb.Product; } }

        public override string Manufacturer { get { return usb.Manufacturer; } }

        public override string SerialNumber { get { return usb.SerialNumber; } }

        public override string ToString() { return DeviceName + " (s/n: " + SerialNumber + ")"; }

        public override float ActualTemperature
        {
            get
            {
                return ReadFloat((int)Cmd.GetTemperature);
            }
        }

        public override int PixelCount { get { return pixelCount; } }
        protected int pixelCount;

        public override float[] Wavelengths { get { return wavelengths; } }
        protected float[] wavelengths;

        public virtual Version SoftwareVersion { get { return softwareVersion; } }
        protected Version softwareVersion;

        public virtual Version HardwareVersion { get { return ReadVersion((int)Cmd.GetHardwareVersion); } }

        //***************** Exposure ********************

        public virtual bool AutoExposure
        {
            get { return autoExposure; }
            set
            {
                WriteInt((int)Cmd2.SetAutoExposureEnable, (value ? 1 : 0));
                autoExposure = value;
            }
        }
        protected bool autoExposure = false;

        public virtual float AutoExposureTime
        {
            get { return autoExposureTime; }
            set
            {
                WriteInt((int)Cmd2.SetAutoExposureTime, (int)Convert.ToUInt32(value * 1E6f)); //Range is checked by device
                autoExposureTime = value;
            }
        }
        protected float autoExposureTime = 0.2f;

        public override float ExposureTime
        {
            get { return exposureTime; }
            set
            {
                WriteInt((int)Cmd.SetExposureTime, (int)Convert.ToUInt32(value * 1E6f)); //Range is checked by device
                exposureTime = value;
            }
        }
        protected float exposureTime;

        public override float MinExposureTime { get { return minExposureTime; } }
        protected float minExposureTime;

        public override float MaxExposureTime { get { return maxExposureTime; } }
        protected float maxExposureTime;

        public override int Averaging
        {
            get { return averaging; }
            set
            {
                if (value != averaging) //Range is checked by device
                {
                    WriteInt((int)Cmd.SetAveraging, value);
                    averaging = value;
                }
            }
        }
        protected int averaging = 1;

        public override int MaxAveraging { get { return maxAveraging; } }
        protected int maxAveraging = 1;

        public override void StartExposure(int NumSpectra)
        {
            UpdateTimeInDevice();
            WriteInt((int)Cmd.StartExposure, NumSpectra);
        }

        public virtual void TakeDarkSpectrum()
        {
            UpdateTimeInDevice();
            Write((int)Cmd2.TakeDarkSpectrum);
        }

        public virtual void TakeReferenceSpectrum()
        {
            UpdateTimeInDevice();
            Write((int)Cmd2.TakeReferenceSpectrum);
        }

        private void UpdateTimeInDevice()
        {
            if (softwareVersion >= minFirmwareDateTime)
            {
                DateTime now = DateTime.Now;
                if (now > nextTimeUpdate)
                {
                    int time = (int)(now.TimeOfDay.TotalMilliseconds * 10);
                    int date = (int)((now.Date - date0).TotalDays);
                    WriteTwoInts((int)Cmd2.SetDateTime, time, date);
                    nextTimeUpdate = now + TimeSpan.FromMinutes(1);
                }
            }
        }
        private DateTime nextTimeUpdate = DateTime.MinValue;

        public override void CancelExposure()
        {
            Write((int)Cmd.CancelExposure);
            if (autoExposure)
            {
                averaging = ReadInt((int)Cmd.GetAveraging);
                exposureTime = ReadInt((int)Cmd.GetExposureTime) * 1E-6f;
            }
        }

        //***************** Status and Read Data ********************

        public override SpectrStatus GetStatus(out int AvailableSpectra)
        {
            if (!IsOpen)
            {
                AvailableSpectra = 0;
                return SpectrStatus.NotConnected;
            }

            int st = ReadInt((int)Cmd.GetStatus);
            AvailableSpectra = (st >> 8) & 0xFFFF;
            st = st & 0xFF;
            if (st < 0) throw new Exception("Device status " + st + " indicates an error.");
            return (SpectrStatus)st;
        }

        public override SpectrumData GetSpectrumData()
        {
            int n = ReadBulk((int)Cmd.GetSpectrum);
            SpectrumData spd = ConvertSpectrumFromDevice(replyBuffer, n);

            if (autoExposure)
            {
                averaging = ReadInt((int)Cmd.GetAveraging);
                exposureTime = ReadInt((int)Cmd.GetExposureTime) * 1E-6f;
            }

            return spd;
        }

        public virtual SpectrumData[] GetAllSpectrumData()
        {
            int st = ReadInt((int)Cmd2.GetAllAvailableSpectra);

            //First reply is same as Cmd.GetStatus (see above)
            int availableSpectra = (st >> 8) & 0xFFFF;
            st = st & 0xFF;
            if (st < 0) throw new Exception("Device status " + st + " indicates an error.");

            //Then all available spectra follow as separate USB transactions
            SpectrumData[] spd = new SpectrumData[availableSpectra];
            for (int i = 0; i < availableSpectra; i++)
            {
                int n = usb.Read(replyBuffer, replyBuffer.Length);
                spd[i] = ConvertSpectrumFromDevice(replyBuffer, n);
            }

            if (autoExposure && availableSpectra > 0)
            {
                averaging = ReadInt((int)Cmd.GetAveraging);
                exposureTime = ReadInt((int)Cmd.GetExposureTime) * 1E-6f;
            }

            return spd;
        }

        protected SpectrumData ConvertSpectrumFromDevice(byte[] buffer, int length) //Data includes return code from device
        {
            if (length < 68) throw new Exception("Only " + length + " bytes in received spectrum.");
            GCHandle handle = GCHandle.Alloc(buffer, GCHandleType.Pinned);
            SpectrumHeader header = (SpectrumHeader)Marshal.PtrToStructure(handle.AddrOfPinnedObject() + 4, typeof(SpectrumHeader)); //replyBuffer includes return code
            handle.Free();

            if (header.PixelCount != pixelCount) throw new Exception("Unexpected PixelCount in received spectrum.");
            if (header.VersionNumber != 0) throw new NotSupportedException("Spectrum data version " + header.VersionNumber + " not supported.");

            SpectrumData spd = new SpectrumData();
            float[] spec = new float[pixelCount];
            spd.Values = spec;
            spd.ExposureTime = header.ExposureTime * 1E-6f;
            spd.Averaging = (int)header.Averaging;
            spd.LoadLevel = header.LoadLevel;
            spd.SensorTemperature = header.SensorTemperature;
            spd.ExposureSettings = header.ExposureSettings;
            //if ((header.ExposureSettings & 0x8000) != 0) System.Diagnostics.Debug.WriteLine("Spectrum dropped in device.");
            spd.AppliedProcessing = (SpectrumData.ProcessingValues)header.AppliedProcessing;
            spd.DarkAvg = header.DarkAvg;
            spd.ReadoutNoise = header.ReadoutNoise;
            int p1 = Array.IndexOf(buffer, (byte)0, 36); //Spectrum name starts at position 32, add 4 for return code
            if (p1 >= 0 && p1 < 66)
            {
                spd.Name = System.Text.Encoding.ASCII.GetString(buffer, 36, p1 - 36);
                p1++;
                int p2 = Array.IndexOf(buffer, (byte)0, p1);
                if (p2 >= 0 && p2 < 66)
                {
                    spd.IntensityUnit = System.Text.Encoding.ASCII.GetString(buffer, p1, p2 - p1);
                    p2++;
                    int p3 = Array.IndexOf(buffer, (byte)0, p2);
                    if (p3 >= 0 && p3 < 66)
                    {
                        spd.Sample = System.Text.Encoding.ASCII.GetString(buffer, p2, p3 - p2);
                    }
                }
            }

            switch (header.DataFormat)
            {
                case 0: //32-bit float values
                    if (length != 4 + 64 + pixelCount * 4) throw new Exception("Unexpected number of bytes " + length + " in received spectrum.");
                    Buffer.BlockCopy(buffer, 68, spec, 0, pixelCount * 4); //68 = size of SpectrumHeader + return code
                    break;
                case 1: //16-bit integer values
                    if (length != 4 + 64 + 8 + pixelCount * 2) throw new Exception("Unexpected number of bytes " + length + " in received spectrum.");
                    float offset = BitConverter.ToSingle(buffer, 68);
                    float slope = BitConverter.ToSingle(buffer, 72);
                    int p2 = 76;
                    for (int p = 0; p < pixelCount; p++)
                        spec[p] = offset + slope * (buffer[p2++] + (buffer[p2++] << 8));
                    break;
                case 2: //16-bit integer values with 2x binning
                    if (length != 4 + 64 + 8 + pixelCount) throw new Exception("Unexpected number of bytes " + length + " in received spectrum.");
                    offset = BitConverter.ToSingle(buffer, 68);
                    slope = BitConverter.ToSingle(buffer, 72);
                    p2 = 76;
                    for (int p = 0; p < pixelCount; p += 2)
                    {
                        float f = offset + slope * (buffer[p2++] + (buffer[p2++] << 8));
                        spec[p] = f;
                        spec[p + 1] = f;
                    }
                    break;
                default:
                    throw new NotSupportedException("Data format " + header.DataFormat + " not supported.");
            }

            if (softwareVersion >= minFirmwareDateTime)
                spd.TimeStamp = date0 + TimeSpan.FromDays(header.Date) + TimeSpan.FromMilliseconds(header.Time * 0.1);
            else //legacy fallback
                spd.TimeStamp = DateTime.Now + TimeSpan.FromMilliseconds(header.Time - ReadInt((int)Cmd.GetSysTick));

            return spd;
        }

        protected byte[] ConvertSpectrumToDevice(SpectrumData spd) //Returned data does not include return code from device
        {
            byte[] data = new byte[64 + spd.PixelCount * 4];
            SpectrumHeader header = new SpectrumHeader();
            header.ExposureTime = Convert.ToUInt32(spd.ExposureTime * 1E6f); //with rounding
            header.Averaging = (uint)spd.Averaging;
            header.LoadLevel = spd.LoadLevel;
            header.SensorTemperature = spd.SensorTemperature;
            header.PixelCount = (ushort)spd.PixelCount;
            header.DataFormat = 0;
            header.VersionNumber = 0;
            header.ExposureSettings = (ushort)spd.ExposureSettings;
            header.AppliedProcessing = (ushort)spd.AppliedProcessing;
            header.DarkAvg = spd.DarkAvg;
            header.ReadoutNoise = spd.ReadoutNoise;
            string combinedName = spd.Name + '\0' + spd.IntensityUnit + '\0' + spd.Sample;
            if (combinedName.Length > SpectrumHeaderNameMaxLength - 1) combinedName = combinedName.Substring(0, SpectrumHeaderNameMaxLength - 1);
            header.Name = combinedName + '\0';
            header.Time = (uint)(spd.TimeStamp.TimeOfDay.TotalMilliseconds * 10);
            header.Date = (uint)((spd.TimeStamp.Date - date0).TotalDays);

            GCHandle handle = GCHandle.Alloc(data, GCHandleType.Pinned);
            Marshal.StructureToPtr(header, handle.AddrOfPinnedObject(), true);
            handle.Free();
            Buffer.BlockCopy(spd.Values, 0, data, 64, spd.PixelCount * 4); //68 = size of SpectrumHeader + return code
            return data;
        }

        public override float[] GetSpectrum()
        {
            SpectrumData spd = GetSpectrumData();
            return spd.Values;
        }

        //***************** Calibration and User Data ********************

        public virtual SpectrumData ReadCalibrSpectrum(CalibrSpectra index)
        {
            if (softwareVersion < minFirmwareCalibrSpectra) throw new Exception("Not currently supported by the device. Please update the firmware.");
            int n = ReadBulk((int)Cmd2.GetOffsetSpectrum + (int)index);
            if (n == 4) return null;
            return ConvertSpectrumFromDevice(replyBuffer, n);
        }

        public virtual void WriteCalibrSpectrum(CalibrSpectra index, SpectrumData spd, int passcode)
        {
            if (softwareVersion < minFirmwareCalibrSpectra) throw new Exception("Not currently supported by the device. Please update the firmware.");
            byte[] data = (spd == null ? new byte[0] : ConvertSpectrumToDevice(spd));
            if (index == CalibrSpectra.OffsetSpectrum || index == CalibrSpectra.FactoryDark || index == CalibrSpectra.FactorySensCal)
                WriteBulk((int)Cmd2.SetOffsetSpectrum + (int)index, passcode, data);
            else
                WriteBulk((int)Cmd2.SetOffsetSpectrum + (int)index, data);
        }

        public virtual void RestoreFactoryCalibration(CalibrSpectra index)
        {
            if (index == CalibrSpectra.DarkSpectrum) Write((int)Cmd2.RestoreFactoryDarkSpectrum);
            else if (index == CalibrSpectra.SensitivityCal) Write((int)Cmd2.RestoreFactorySensitivityCal);
            else throw new ArgumentException();
        }

        public virtual byte[] ReadUserData()
        {
            int n = ReadBulk((int)Cmd.GetUserData) - 4;
            if (n == 0) throw new InvalidOperationException("No user data stored in device.");
            byte[] res = new byte[n];
            Array.Copy(replyBuffer, 4, res, 0, n);
            return res;
        }

        public virtual void WriteUserData(byte[] data)
        {
            int maxbytes = ReadInt((int)Cmd2.GetUserDataMaxSize);
            if (data.Length > maxbytes) throw new ArgumentException("User data exceeds the maximum size of " + maxbytes + " bytes.");
            WriteBulk((int)Cmd.SetUserData, data);
        }

        public virtual bool SubtractDark
        {
            get { return subtractDark; }
            set
            {
                WriteInt((int)Cmd2.SetSubtractDark, (value ? 1 : 0));
                subtractDark = value;
            }
        }
        protected bool subtractDark;

        public virtual SpectrIntensityUnits IntensityUnit
        {
            get { return intensityUnit; }
            set
            {
                WriteInt((int)Cmd2.SetIntensityUnit, (int)value);
                intensityUnit = value;
            }
        }
        protected SpectrIntensityUnits intensityUnit;

        //***************** Basic communication protocol ********************

        protected byte[] replyBuffer = new byte[16492];

        protected void Write(int cmd)
        {
            SendReceiveMessage(new byte[] { (byte)cmd, (byte)(cmd >> 8), 0, 0 }, replyBuffer);
        }

        public void WriteInt(int cmd, int value)
        {
            byte[] bb = new byte[8];
            CopyIntToByteArray(cmd, bb, 0);
            CopyIntToByteArray(value, bb, 4);
            SendReceiveMessage(bb, replyBuffer);
        }

        public void WriteTwoInts(int cmd, int value1, int value2)
        {
            byte[] bb = new byte[12];
            CopyIntToByteArray(cmd, bb, 0);
            CopyIntToByteArray(value1, bb, 4);
            CopyIntToByteArray(value2, bb, 8);
            SendReceiveMessage(bb, replyBuffer);
        }

        private void WriteFloat(int cmd, float value)
        {
            byte[] bb = new byte[8];
            CopyIntToByteArray(cmd, bb, 0);
            byte[] ff = BitConverter.GetBytes(value);
            Buffer.BlockCopy(ff, 0, bb, 4, 4);
            SendReceiveMessage(bb, replyBuffer);
        }

        public string ReadString(int cmd)
        {
            SendReceiveMessage(new byte[] { (byte)cmd, (byte)(cmd >> 8), 0, 0 }, replyBuffer);
            int p = Array.IndexOf<byte>(replyBuffer, 0, 4);
            if (p < 0) p = replyBuffer.Length;
            return Encoding.ASCII.GetString(replyBuffer, 4, p - 4);
        }

        public int ReadInt(int cmd)
        {
            SendReceiveMessage(new byte[] { (byte)cmd, (byte)(cmd >> 8), 0, 0 }, replyBuffer);
            return BitConverter.ToInt32(replyBuffer, 4);
        }

        protected float ReadFloat(int cmd)
        {
            SendReceiveMessage(new byte[] { (byte)cmd, (byte)(cmd >> 8), 0, 0 }, replyBuffer);
            return BitConverter.ToSingle(replyBuffer, 4);
        }

        protected Version ReadVersion(int cmd)
        {
            SendReceiveMessage(new byte[] { (byte)cmd, (byte)(cmd >> 8), 0, 0 }, replyBuffer);
            return new Version(replyBuffer[7], replyBuffer[6], replyBuffer[5], replyBuffer[4]); //little-endian
        }

        protected int ReadBulk(int cmd) //Returns number of bytes received. Data is in private field replyBuffer and will be reused. Includes return code.
        {
            int n = SendReceiveMessage(new byte[] { (byte)cmd, (byte)(cmd >> 8), 0, 0 }, replyBuffer);
            return n;
        }

        protected int ReadBulk(int cmd, int page) //Returns number of bytes received. Data is in private field replyBuffer and will be reused. Includes return code.
        {
            byte[] bb = new byte[8];
            CopyIntToByteArray(cmd, bb, 0);
            CopyIntToByteArray(page, bb, 4);
            int n = SendReceiveMessage(bb, replyBuffer);
            return n;
        }

        protected void WriteBulk(int cmd, byte[] data)
        {
            byte[] bb = new byte[data.Length + 4];
            CopyIntToByteArray(cmd, bb, 0);
            Array.Copy(data, 0, bb, 4, data.Length);
            SendReceiveMessage(bb, replyBuffer);
        }

        protected void WriteBulk(int cmd, int param, byte[] data)
        {
            byte[] bb = new byte[data.Length + 8];
            CopyIntToByteArray(cmd, bb, 0);
            CopyIntToByteArray(param, bb, 4);
            Array.Copy(data, 0, bb, 8, data.Length);
            SendReceiveMessage(bb, replyBuffer);
        }

        protected void CopyIntToByteArray(int integer, byte[] array, int byteOffset)
        {
            array[byteOffset++] = (byte)integer;
            array[byteOffset++] = (byte)(integer >> 8);
            array[byteOffset++] = (byte)(integer >> 16);
            array[byteOffset] = (byte)(integer >> 24);
        }

        public long stopWatchTicks;

        public virtual int SendReceiveMessage(byte[] sendMessage, byte[] replyBuffer) //returns number of bytes received (including return code)
        {
            if (usb == null || !usb.IsOpen) throw new Exception("Interface is closed.");
            
            usb.Write(sendMessage, sendMessage.Length);

            System.Diagnostics.Stopwatch sw = System.Diagnostics.Stopwatch.StartNew();
            int len = usb.Read(replyBuffer, replyBuffer.Length);
            stopWatchTicks = sw.ElapsedTicks;

            if (len == 0) throw new CommunicationException("No reply from device.");
            if (len < 4) throw new CommunicationException("Not enough bytes received from device.");
            int retcode = BitConverter.ToInt32(replyBuffer, 0);
            if ((retcode >> 8) != 0) //if error was reported
            {
                string s = " (cmd:" + BitConverter.ToInt32(sendMessage, 0).ToString("X4") + " err:" + retcode.ToString("X4") + ").";
                switch (retcode >> 8)
                {
                    case 1: throw new ArgumentException("Unknown command code" + s);
                    case 2: throw new ArgumentException("Invalid parameter sent to device" + s);
                    case 4: throw new InvalidOperationException("This operation is currently not allowed" + s);
                    case 5: throw new NotSupportedException("The device does not support this operation" + s);
                    case 6: throw new ArgumentException("Invalid passcode for this operation" + s);
                    case 7: throw new CommunicationException("A communication error occurred" + s);
                    case 8: throw new Exception("Internal error received from device" + s);
                    case 10: throw new Exception("Cannot read from device memory (" + (retcode == 0x0A01 ? "checksum error" : "is empty") + ").");
                    case 11: throw new Exception("Spectrum processing error: " + ((retcode & 0xFF) < ProcErrMessages.Length ? ProcErrMessages[retcode & 0xFF] : s));
                    default: throw new Exception("Error code 0x" + retcode.ToString("X4") + " received from device" + s);
                }
            }
            if ((sendMessage[0] == 0x02 || sendMessage[0] == 0xFF) && sendMessage[1] == 0x00
                && sendMessage[2] == 0x00 && sendMessage[3] == 0x00) usb.Close(); //Command DeviceReset or EnterBootloader
            return len; //normal case
        }

        private string[] ProcErrMessages = new string[] { "", "Pixel count mismatch", "Dark spectrum missing",
            "Reference spectrum missing", "Sensitivity calibration missing" };

        [StructLayout(LayoutKind.Sequential, Pack = 1, CharSet = CharSet.Ansi)]
        private struct SpectrumHeader
        {
            public uint ExposureTime; //in us (0: unknown)
            public uint Averaging; //(0: unknown)
            public uint Time; //elapsed 0.1 ms since midnight
            public uint Date; //elapsed days
            public float LoadLevel; //1.0: Saturation Level (< 0: unknown or n/a)
            public float SensorTemperature; //in °C (< 274: unknown)
            public ushort PixelCount;
            public byte DataFormat;
            public byte VersionNumber; 
            public ushort ExposureSettings; 
            public ushort AppliedProcessing; 
            [MarshalAs(UnmanagedType.ByValTStr, SizeConst = SpectrumHeaderNameMaxLength)]
            public string Name; //up to 3 null-terminated strings: Spectrum name, intensity unit, sample name
            public float DarkAvg;
            public float ReadoutNoise;
        }

        private const int SpectrumHeaderNameMaxLength = 24; //if you change this number, also change number in writeCalibration_Click()

        protected enum Cmd //old command codes
        {
            Reset = 0x0000,
            StandBy = 0x0001,
            StartExposure = 0x0004,
            CancelExposure = 0x005,

            GetExposureTime = 0x1000,
            SetExposureTime = 0x1100,
            GetMinExposureTime = 0x1200,
            GetMaxExposureTime = 0x1300,
            GetAveraging = 0x1001,
            SetAveraging = 0x1101,
            GetMaxAveraging = 0x1301,

            GetHardwareVersion = 0x2004,
            GetSoftwareVersion = 0x2005,
            GetPixelCount = 0x2007,

            GetStatus = 0x3000,
            GetTemperature = 0x3001,
            GetSysTick = 0x3003,

            GetSpectrum = 0x4000,
            GetWavelengths = 0x4001,
            GetWavelengthCoefficients = 0x4005,

            GetUserData = 0x4003,
            SetUserData = 0x4103
        }

        protected enum Cmd2 //new command codes
        {
            TakeDarkSpectrum = 0x0006,
            TakeReferenceSpectrum = 0x0007,
            RestoreFactoryDarkSpectrum = 0x0008,
            RestoreFactorySensitivityCal = 0x0009,
            GetAllAvailableSpectra = 0x000A,
            GetAutoExposureTime = 0x100A,
            SetAutoExposureTime = 0x110A,
            GetAutoExposureEnable = 0x1009,
            SetAutoExposureEnable = 0x1109,
            GetSubtractDark = 0x100E,
            SetSubtractDark = 0x110E,
            GetIntensityUnit = 0x100F,
            SetIntensityUnit = 0x110F,
            GetWavelengthCoeff0 = 0x201C,
            GetUserDataMaxSize = 0x2014,
            GetSupplyVoltage = 0x3008,
            EnterBootloader = 0x00FF,
            GetDateTime = 0x4005,
            SetDateTime = 0x4105,
            GetOffsetSpectrum = 0x4006,
            SetOffsetSpectrum = 0x4106,
            GetTestZeroLengthPacket = 0x400C,
            SetTestZeroLengthPacket = 0x410C
        }

        public enum CalibrSpectra
        {
            OffsetSpectrum = 0,
            FactoryDark = 1,
            FactorySensCal = 2,
            DarkSpectrum = 3,
            SensitivityCal = 4,
            RefSpectrum = 5
        }
    }
}
