﻿using DeviceClasses;
using System;
using System.Drawing;
using System.IO;
using System.Windows.Forms;
using System.Globalization;
using System.Diagnostics;
using System.Text;

namespace FastReadout
{
    public partial class MainForm : Form
    {
        private Spectrometer spectr;
        bool updatingUI = false;
        int dataFormat = 0;

        public MainForm()
        {
            InitializeComponent();
            deviceNameLabel.Text = "";
            deviceSerialNoLabel.Text = "";
        }

        private void openButton_Click(object sender, EventArgs e)
        {
            Spectrometer[] devices = ZioLinkSpectrometer.FindDevices();
            if (devices.Length == 0)
                deviceLabel.Text = "No device found.";
            else
            {
                spectr = devices[0]; //only consider first device

                updatingUI = true;
                try
                {
                    spectr.Open();

                    expTimeBox.Maximum = (decimal)(spectr.MaxExposureTime * 1000f);
                    expTimeBox.Minimum = (decimal)(spectr.MinExposureTime * 1000f);
                    expTimeBox.Value = (decimal)(spectr.ExposureTime * 1000f);
                    averagingBox.Maximum = (decimal)spectr.MaxAveraging;
                    averagingBox.Value = spectr.Averaging;

                    if (spectr is ZioLinkSpectrometer)
                    {
                        (spectr as ZioLinkSpectrometer).AutoExposure = false;
                        if ((spectr as ZioLinkSpectrometer).SoftwareVersion >= new Version(4, 2, 0, 0))
                        {
                            dataFormat = (spectr as ZioLinkSpectrometer).ReadInt(0x1014); //0x1014 = Get Parameter DataFormat
                            switch (dataFormat)
                            {
                                case 0: dataFormat32BitRadioButton.Checked = true; break;
                                case 1: dataFormat16BitRadioButton.Checked = true; break;
                                case 2: dataFormat16BitBinningRadioButton.Checked = true; break;
                            }
                        }
                    }

                    deviceLabel.Text = "Device found:";
                    deviceNameLabel.Text = spectr.DeviceName;
                    deviceSerialNoLabel.Text = "s/n: " + spectr.SerialNumber;
                    openButton.Enabled = false;
                    InProgress = false; //update button states
                }
                catch (Exception ex)
                {
                    deviceLabel.Text = ex.Message;
                    MessageBox.Show(ex.Message, "Cannot open spectrometer", MessageBoxButtons.OK, MessageBoxIcon.Error);
                }
                updatingUI = false;
            }
        }

        private void expTimeBox_ValueChanged(object sender, EventArgs e)
        {
            if (updatingUI) return; //if value is changed programatically, not by user action
            try
            {
                spectr.ExposureTime = (float)expTimeBox.Value * 0.001f;
            }
            catch (Exception ex)
            {
                MessageBox.Show(ex.Message, "Cannot set exposure time", MessageBoxButtons.OK, MessageBoxIcon.Error);
            }
        }

        private void averagingBox_ValueChanged(object sender, EventArgs e)
        {
            if (updatingUI) return; //if value is changed programatically, not by user action
            try
            {
                spectr.Averaging = (int)averagingBox.Value;
            }
            catch (Exception ex)
            {
                MessageBox.Show(ex.Message, "Cannot set averaging", MessageBoxButtons.OK, MessageBoxIcon.Error);
            }
        }

        private void dataFormat_Clicked(object sender, EventArgs e)
        {
            try
            {
                if (dataFormat32BitRadioButton.Checked) (spectr as ZioLinkSpectrometer).WriteInt(0x1114, 0); //0x1114 = Set Parameter DataFormat
                else if (dataFormat16BitRadioButton.Checked) (spectr as ZioLinkSpectrometer).WriteInt(0x1114, 1);
                else if (dataFormat16BitBinningRadioButton.Checked) (spectr as ZioLinkSpectrometer).WriteInt(0x1114, 2);
            }
            catch (Exception ex)
            {
                MessageBox.Show(ex.Message, "Cannot set data format", MessageBoxButtons.OK, MessageBoxIcon.Error);
            }
        }

        private void startButton_Click(object sender, EventArgs e)
        {
            ClearLog();
            InProgress = true;
            int numSpectra = (int)numSpectraBox.Value;
            int remainingSpectra = numSpectra;
            double maxTime = 0;
            double minTime = double.MaxValue;
            double avgTime = 0;
            bool anyDropped = false;
            stopWatch = System.Diagnostics.Stopwatch.StartNew();
            appendLog("Starting.");
            try
            {
                spectr.StartExposure(remainingSpectra);
                appendLog("Taking " + remainingSpectra + " spectra.");

                while (remainingSpectra > 0)
                {
                    if (getAllAtOnceCheckBox.Checked)
                    {
                        System.Diagnostics.Stopwatch stopWatch2 = System.Diagnostics.Stopwatch.StartNew();
                        SpectrumData[] spd = (spectr as ZioLinkSpectrometer).GetAllSpectrumData();
                        double time = elapsedTime_ms(stopWatch2.ElapsedTicks);

                        if (spd.Length == 0)
                        {
                            updateLog();
                            Application.DoEvents(); //no spectrum available: we have some time for UI tasks
                            if (stopButtonClicked)
                            {
                                appendLog("Exposure cancelled by user.");
                                return;
                            }
                        }
                        else
                        {
                            bool bufferFull = false;
                            for (int i = 0; i < spd.Length; i++)
                            {
                                if ((spd[i].ExposureSettings & 0x8000) != 0) //this is how the spectrometer reports that its internal buffer was full
                                {
                                    bufferFull = true;
                                    anyDropped = true;
                                }
                            }

                            time = time / spd.Length; //average time per spectrum
                            if (spd.Length == 1) appendLog("1  new spectrum read in " + time.ToString("0.00") + " ms" + (bufferFull ? " (spectrum dropped)." : "."));
                            else appendLog(spd.Length.ToString() + " new spectra read in " + time.ToString("0.00") + " ms/spectrum" + (bufferFull ? " (spectrum dropped)." : "."));

                            if (time > maxTime) maxTime = time;
                            if (time < minTime) minTime = time;
                            avgTime += (time * spd.Length); //need to add time for each spectrum
                            remainingSpectra -= spd.Length;
                        }
                    }
                    else
                    {
                        SpectrStatus st = spectr.GetStatus(out int available);

                        if (st != SpectrStatus.TakingSpectrum && (remainingSpectra - available) > 0)
                            throw new Exception("Spectrometer should be taking spectra, but status is: " + statusStr[(int)st]);

                        if (available > 1) stopButton.Visible = false; //in this case there probably won't be time to do Application.DoEvents() anyway

                        if (available == 0)
                        {
                            updateLog();
                            Application.DoEvents(); //no spectrum available: we have some time for UI tasks
                            if (stopButtonClicked)
                            {
                                appendLog("Exposure cancelled by user.");
                                return;
                            }
                        }
                        else
                        {
                            appendLog(available.ToString() + (available == 1 ? " new spectrum available." : " new spectra available."));

                            for (int i = 0; i < available; i++)
                            {
                                System.Diagnostics.Stopwatch stopWatch2 = System.Diagnostics.Stopwatch.StartNew();
                                SpectrumData sp = spectr.GetSpectrumData();
                                double time = elapsedTime_ms(stopWatch2.ElapsedTicks);
                                bool bufferFull = ((sp.ExposureSettings & 0x8000) != 0); //this is how the spectrometer reports that its internal buffer was full
                                if (bufferFull) anyDropped = true;
                                appendLog("Spectrum read in " + time.ToString("0.00") + " ms" + (bufferFull ? " (spectrum dropped)." : ".")
                                    + "   " + elapsedTime_ms((spectr as ZioLinkSpectrometer).stopWatchTicks).ToString("0.00"));
                                if (time > maxTime) maxTime = time;
                                if (time < minTime) minTime = time;
                                avgTime += time;
                                remainingSpectra--;
                            }
                        }
                    }
                }
                double totalTime = elapsedTime_ms(stopWatch.ElapsedTicks);
                appendLog("All spectra received from device.");
                stopWatch = null;

                appendLog("");
                appendLog(numSpectra.ToString() + " spectra taken in " + totalTime.ToString("0.00") + " ms (" + (totalTime / numSpectra).ToString("0.00") + " ms per spectrum).");
                appendLog("Spectrum transfer time: avg:" + (avgTime / numSpectra).ToString("0.00") + "  min:" + minTime.ToString("0.00") + "  max:" + maxTime.ToString("0.00") + " ms.");
                appendLog(anyDropped ? "Spectra were not taken continously, because the device buffer went full." : "All spectra were taken continuously.");
            }
            catch (Exception ex)
            {
                appendLog("ERROR: " + ex.Message);
                MessageBox.Show(ex.Message, "Error During Measurement", MessageBoxButtons.OK, MessageBoxIcon.Error);
            }
            finally
            {
                updateLog();
                InProgress = false;
            }
        }

        System.Diagnostics.Stopwatch stopWatch = null;

        private double elapsedTime_ms(long ticks)
        {
            return (double)ticks / (double)System.Diagnostics.Stopwatch.Frequency * 1000;
        }

        private bool InProgress
        {
            get
            {
                return inProgress;
            }
            set
            {
                inProgress = value;
                startButton.Enabled = !inProgress;

                stopButton.Visible = inProgress;
                stopButton.Enabled = true;
                stopButtonClicked = false;

                expTimeBox.Enabled = !inProgress;
                averagingBox.Enabled = !inProgress;
                numSpectraBox.Enabled = !inProgress;

                if (spectr is ZioLinkSpectrometer)
                {
                    if ((spectr as ZioLinkSpectrometer).SoftwareVersion >= new Version(4, 2, 0, 0))
                    {
                        dataFormat32BitRadioButton.Enabled = !inProgress;
                        dataFormat16BitRadioButton.Enabled = !inProgress;
                        dataFormat16BitBinningRadioButton.Enabled = !inProgress;
                        dataFormat16BitBinningLabel.Enabled = !inProgress;
                        getAllAtOnceCheckBox.Enabled = !inProgress;
                    }
                }
            }
        }
        private bool inProgress = false;

        private void appendLog(string text)
        {
            if (stopWatch == null)
                logBuffer.AppendLine(text);
            else
                logBuffer.AppendLine(elapsedTime_ms(stopWatch.ElapsedTicks).ToString("00000.00") + " ms: " + text);
        }
        StringBuilder logBuffer = new StringBuilder(100000);

        private void updateLog()
        {
            logBox.AppendText(logBuffer.ToString());
            logBuffer.Clear();
        }

        private void ClearLog()
        {
            logBox.Clear();
            logBuffer.Clear();
        }


        /*private void readoutTimer_Tick(object sender, EventArgs e)
        {
            try
            {

            }
            catch (Exception ex)
            {
                readoutTimer.Enabled = false;
                MessageBox.Show(ex.Message, "Cannot read spectrum from device", MessageBoxButtons.OK, MessageBoxIcon.Error);
            }
        }*/

        private string[] statusStr = new string[] { "Idle", "Waiting for trigger", "Taking spectrum",
            "Waiting for temperature", "?", "?", "?", "?", "Powered Off", "Sleep Mode", "" };

        private void closeButton_Click(object sender, EventArgs e)
        {
            Close();
        }

        private void MainForm_FormClosing(object sender, FormClosingEventArgs e)
        {
            if (inProgress) e.Cancel = true;
        }

        private void MainForm_FormClosed(object sender, FormClosedEventArgs e)
        {
            if (spectr != null)
            {
                try
                {
                    if (spectr is ZioLinkSpectrometer)
                        (spectr as ZioLinkSpectrometer).WriteInt(0x1114, dataFormat); //restore previous setting for this parameter
                }
                catch (Exception ex) { System.Diagnostics.Debug.WriteLine(ex.Message + " while closing device connection."); }
                spectr.Close(); //does not throw exceptions
            }
        }

        private void stopButton_Click(object sender, EventArgs e)
        {
            stopButtonClicked = true;
            stopButton.Enabled = false; //give immediate visual feedback
            stopButton.Update();
        }
        private bool stopButtonClicked = false;
    }
}
