﻿using System;
using System.Collections.Generic;
using System.Drawing;
using System.Drawing.Drawing2D;
using System.ComponentModel;
using System.Linq;
using System.Text;
using System.Threading.Tasks;
using System.Windows.Forms;

namespace AuxControls
{
    public class Diagram : PictureBox
    {
        //****** Diagram Layers ******

        public enum CurveStyle { Auto = 0, Lines, Points, LinesAndPoints }

        public const int maxZPosition = 3;

        public class LayerBaseClass : Object
        {
            public bool Visible = true;
            public int ZPosition = 1; //so far: 0 (background), 1 (default), 2 (foreground), 3 (overlay text)
            public int ID = 0;
        }

        public class SinglePrecDataLayer : LayerBaseClass
        {
            public CurveStyle Style;
            public Color Color;
            public float[] DataX;
            public float[] DataY;

            public SinglePrecDataLayer(float[] dataX, float[] dataY, Color color, CurveStyle style)
            { this.DataX = dataX; this.DataY = dataY; this.Color = color; this.Style = style; }
        }

        public class DoublePrecDataLayer : LayerBaseClass
        {
            public CurveStyle Style;
            public Color Color;
            public double[] DataX;
            public double[] DataY;

            public DoublePrecDataLayer(double[] dataX, double[] dataY, Color color, CurveStyle style)
            { this.DataX = dataX; this.DataY = dataY; this.Color = color; this.Style = style; }
        }

        public class HorizontalSelectionLayer : LayerBaseClass
        {
            public double SelectionStart;
            public double SelectionEnd;

            public HorizontalSelectionLayer(double start, double end)
            { this.SelectionStart = start; this.SelectionEnd = end; }
        }

        public class HorizontalLineLayer : LayerBaseClass
        {
            public double Position;
            public Color Color;

            public HorizontalLineLayer(double position, Color color)
            { this.Position = position; this.Color = color; }
        }

        public class VerticalLineLayer : LayerBaseClass
        {
            public double Position;
            public Color Color;

            public VerticalLineLayer(double position, Color color)
            { this.Position = position; this.Color = color; }
        }

        public class DataLabelLayer : LayerBaseClass
        {
            public Color Color;
            public float[] DataX;
            public float[] DataY;
            public string Text;
        }

        public class WindowLabelLayer : LayerBaseClass
        {
            public Color Color;
            public ContentAlignment Alignment;
            public float Line;
            public string Text;

            public WindowLabelLayer(string text, ContentAlignment alignment, float line, Color color)
            { this.Text = text; this.Alignment = alignment; this.Line = line; this.Color = color; }
        }

        public class CustomGraphicsLayer : LayerBaseClass
        {
            public PaintEventHandler CustomPaint;

            public CustomGraphicsLayer(PaintEventHandler customPaint)
            { this.CustomPaint = customPaint; }
        }

        //Remember to call Invalidate() after change anything in this list in order to repaint the control.
        public List<LayerBaseClass> Layers { get => layers; }
        private List<LayerBaseClass> layers = new List<LayerBaseClass>();

        public int BorderTop { get { return borderTop; } set { borderTop = value; transformValid = false; Invalidate(); } }
        private int borderTop = 10;

        public int BorderBottom { get { return borderBottom; } set { borderBottom = value; transformValid = false; Invalidate(); } }
        private int borderBottom = 40;

        public int BorderLeft { get { return borderLeft; } set { borderLeft = value; transformValid = false; Invalidate(); } }
        private int borderLeft = 40;

        public int BorderRight { get { return borderRight; } set { borderRight = value; transformValid = false; Invalidate(); } }
        private int borderRight = 10;

        public float MinX { get { return minX; } set { minX = value; transformValid = false; Invalidate(); } }
        private float minX = -1;

        public float MaxX { get { return maxX; } set { maxX = value; transformValid = false; Invalidate(); } }
        private float maxX = 1;

        public float MinY { get { return minY; } set { minY = value; transformValid = false; Invalidate(); } }
        private float minY = -1;

        public float MaxY { get { return maxY; } set { maxY = value; transformValid = false; Invalidate(); } }
        private float maxY = 1;

        public bool LogX { get { return logX; } set { logX = value; transformValid = false; Invalidate(); } }
        private bool logX = false;

        public bool LogY { get { return logY; } set { logY = value; transformValid = false; Invalidate(); } }
        private bool logY = false;

        public bool ShowGrid { get => showGrid; set { showGrid = value; Invalidate(); } }
        private bool showGrid = false;

        public Color DiagramColor { get => diagramColor; set { diagramColor = value; diagramBrush = new SolidBrush(value); Invalidate(); } }
        private Color diagramColor = Color.White;
        private Brush diagramBrush = new SolidBrush(Color.White);

        [Browsable(true)]
        public override Color ForeColor { get => base.ForeColor; set { base.ForeColor = value; axisPen = new Pen(value); axisBrush = new SolidBrush(value); Invalidate(); } }
        private Pen axisPen;
        private Brush axisBrush;

        [Browsable(true)]
        public Color GridColor { get => gridColor; set { gridColor = value; gridPen = new Pen(value); Invalidate(); } }
        private Color gridColor = Color.LightGray;
        private Pen gridPen = new Pen(Color.LightGray);

        [Browsable(true)]
        public Color SelectionColor { get => selectionColor; set { selectionColor = value; selectionBrush = new SolidBrush(value); Invalidate(); } }
        private Color selectionColor = Color.FromArgb(212, 212, 212);
        private Brush selectionBrush = new SolidBrush(Color.FromArgb(212, 212, 212));

        [Browsable(true)]
        public override Font Font { get => base.Font; set { base.Font = value; transformValid = false; Invalidate(); } }

        [Browsable(true)]
        public string AxisLabelX { get => axisLabelX; set { axisLabelX = value; Invalidate(); } }
        private string axisLabelX = "";

        [Browsable(true)]
        public string AxisLabelY { get => axisLabelY; set { axisLabelY = value; Invalidate(); } }
        private string axisLabelY = "";

        public Diagram()
        {
            ForeColor = ForeColor; //set axisPen and axisBrush with default color from parent class
        }

        //*********** coordinate transformation *************

        private bool transformValid = false;
        private int offsetX, offsetY;
        private int width1, height1;
        private int canvasWidth, canvasHeight;
        private float[] axisTicksX;
        private float[] axisTicksY;
        private float dx, dy, m11, m22;
        private Rectangle canvasArea;

        private void updateTransform(Graphics gr)
        {
            offsetX = borderLeft;
            offsetY = borderTop;
            width1 = this.Width - 1;
            height1 = this.Height - 1;
            canvasWidth = this.Width - offsetX - borderRight;
            canvasHeight = this.Height - offsetY - borderBottom;
            canvasArea = new Rectangle(offsetX + 1, offsetY + 1, canvasWidth - 1, canvasHeight - 1);

            if (minX >= maxX) return;
            if (minY >= maxY) return;
            if (canvasWidth <= 0) return;
            if (canvasHeight <= 0) return;
            if (logX && minX <= 0) return;
            if (logY && minY <= 0) return;

            m11 = (this.Width - borderLeft - borderRight) / (maxX - minX);
            m22 = -(this.Height - borderTop - borderBottom) / (maxY - minY);
            dx = borderLeft - minX * m11;
            dy = this.Height - borderBottom - minY * m22;
            if (m11 <= 0 || float.IsInfinity(m11)) return;
            if (m22 >= 0 || float.IsInfinity(m22)) return;

            SizeF sz = gr.MeasureString("0.00E0", this.Font);
            axisTicksX = getAxisTicks(sz.Width / m11, minX, maxX, logX);
            axisTicksY = getAxisTicks(sz.Height / (-m22), minY, maxY, logY);

            transformValid = true;
        }

        private float[] axisStepDown = new float[] { 1f, 0.5f, 0.2f, 0.1f, 0.05f, 0.02f, 0.01f, 0.005f, 0.002f };

        private float[] getAxisTicks(float mindist, float min, float max, bool log)
        {
            if (log)
            {
                if (min < 0) throw new ArgumentException("Axis must not be negative for logarithmic scale.");
                int start = (int)Math.Ceiling(Math.Log10(min));
                int n = (int)Math.Log10(max) - start + 1;
                float[] ticks = new float[n];
                for (int i = 0; i < n; i++) ticks[i] = (float)Math.Pow(10, start + i);
                return ticks;
            }
            else
            {
                float step = (float)Math.Pow(10, Math.Ceiling(Math.Log10(max - min)));
                int i;
                for (i = 1; i < axisStepDown.Length; i++)
                {
                    if (step * axisStepDown[i - 1] > (max - min) / 2) continue; //if less than two step widths fit into range: try shorter step (checking previous candidate, because that's the one used below)
                    if (step * axisStepDown[i] < mindist) break; //if axis ticks are getting too close: stop
                    if (step * axisStepDown[i] < (max - min) / 20) break; //use not more than 20 axis ticks in total
                }
                step = step * axisStepDown[i - 1];

                float start = (float)Math.Ceiling(min / step) * step;
                int n = (int)((max - start) / step) + 1;
                float[] ticks = new float[n];
                for (i = 0; i < n; i++) ticks[i] = start + i * step;
                return ticks;
            }
        }

        public float coordToCanvasX(float xd)
        {
            float xc;
            if (logX)
            {
                if (xd <= 0)
                    return 0; //left edge of control
                else
                    xc = (float)((Math.Log10(xd) - Math.Log10(minX)) / (Math.Log10(maxX) - Math.Log10(minX)) * (this.Width - borderLeft - borderRight) + borderLeft);
            }
            else
            {
                xc = xd * m11 + dx;
            }
            return xc;
        }

        public float coordToCanvasY(float yd)
        {
            float yc;
            if (logY)
            {
                if (yd <= 0)
                    return height1; //bottom edge of control
                else
                    yc = (float)((1 - (Math.Log10(yd) - Math.Log10(minY)) / (Math.Log10(maxY) - Math.Log10(minY))) * (this.Height - borderTop - borderBottom) + borderTop);
            }
            else
            {
                yc = yd * m22 + dy;
            }
            return yc;
        }

        public float canvasToCoordX(float xc)
        {
            if (logX)
                return (float)Math.Pow(10, ((xc - borderLeft) / (this.Width - borderLeft - borderRight)) * (Math.Log10(maxX) - Math.Log10(minX)) + Math.Log10(minX));
            else
                return (xc - dx) / m11;
        }

        public float canvasToCoordY(float yc)
        {
            if (logY)
                return (float)Math.Pow(10, (1 - (yc - borderTop) / (this.Height - borderTop - borderBottom)) * (Math.Log10(maxY) - Math.Log10(minY)) + Math.Log10(minY));
            else
                return (yc - dy) / m22;
        }

        // *****************************

        protected override void OnPaint(PaintEventArgs pe)
        {
            base.OnPaint(pe);

            Graphics gr = pe.Graphics;
            Region oldclip = gr.Clip;

            try
            {
                if (!transformValid) updateTransform(gr);

                Rectangle canvasClip = pe.ClipRectangle;
                canvasClip.Intersect(canvasArea); //reduce clip to intersection with canvasArea
                bool needToDrawAxes = !canvasArea.Contains(pe.ClipRectangle);
                bool needToDrawCanvas = canvasArea.IntersectsWith(pe.ClipRectangle);

                if (needToDrawCanvas)
                    gr.FillRectangle(diagramBrush, offsetX, offsetY, canvasWidth, canvasHeight);

                if (layers.Count == 1 && layers[0] is CustomGraphicsLayer)
                {
                    (layers[0] as CustomGraphicsLayer).CustomPaint(this, new PaintEventArgs(gr, new Rectangle(borderLeft, borderTop, canvasWidth, canvasHeight)));
                }
                else if (layers.Count != 0)
                {
                    if (!transformValid) //if still not valid
                    {
                        gr.DrawString("Invalid diagram coordinates.", this.Font, this.axisBrush, borderLeft + 10, borderTop + 10);
                        return;
                    }
                    float xc, yc, f;

                    if (needToDrawCanvas)
                    {                 
                        gr.SetClip(canvasClip); //intersection of ClipRectangle and canvasArea

                        if (showGrid)
                        {
                            for (int i = 0; i < axisTicksX.Length; i++)
                            {
                                xc = coordToCanvasX(axisTicksX[i]);
                                gr.DrawLine(gridPen, xc, borderTop, xc, this.Height - borderBottom);
                            }
                            for (int i = 0; i < axisTicksY.Length; i++)
                            {
                                yc = coordToCanvasY(axisTicksY[i]);
                                gr.DrawLine(gridPen, borderLeft, yc, this.Width - borderRight, yc);
                            }
                        }

                        object previousXArray = null;
                        Point[] pts = new Point[0];
                        for (int z = 0; z <= maxZPosition; z++)
                        {
                            for (int i = 0; i < layers.Count; i++)
                            {
                                if (layers[i].Visible && layers[i].ZPosition == z)
                                {
                                    if (layers[i] is SinglePrecDataLayer)
                                    {
                                        SinglePrecDataLayer ly = layers[i] as SinglePrecDataLayer;
                                        if (previousXArray != ly.DataX)
                                        {
                                            if (pts.Length != ly.DataX.Length) pts = new Point[ly.DataX.Length];
                                            for (int j = 0; j < pts.Length; j++) pts[j].X = (int)(coordToCanvasX(ly.DataX[j]) + 0.5f);
                                            previousXArray = ly.DataX;
                                        }
                                        for (int j = 0; j < pts.Length; j++) pts[j].Y = (int)(coordToCanvasY(ly.DataY[j]) + 0.5f);

                                        if (ly.Style == CurveStyle.Lines || ly.Style == CurveStyle.LinesAndPoints || ly.Style == CurveStyle.Auto)
                                            gr.DrawLines(new Pen(ly.Color), pts);
                                        if (ly.Style == CurveStyle.Points || ly.Style == CurveStyle.LinesAndPoints || (ly.Style == CurveStyle.Auto && pts.Length <= 20))
                                        {
                                            Brush br = new SolidBrush(ly.Color);
                                            for (int j = 0; j < pts.Length; i++)
                                                gr.FillRectangle(br, pts[j].X - 2, pts[j].Y - 2, 5, 5);
                                        }
                                    }
                                    else if (layers[i] is DoublePrecDataLayer)
                                    {
                                        DoublePrecDataLayer ly = layers[i] as DoublePrecDataLayer;
                                        if (previousXArray != ly.DataX)
                                        {
                                            if (pts.Length != ly.DataX.Length) pts = new Point[ly.DataX.Length];
                                            for (int j = 0; j < pts.Length; j++) pts[j].X = (int)(coordToCanvasX((float)ly.DataX[j]) + 0.5f);
                                            previousXArray = ly.DataX;
                                        }
                                        for (int j = 0; j < pts.Length; j++) pts[j].Y = (int)(coordToCanvasY((float)ly.DataY[j]) + 0.5f);

                                        if (ly.Style == CurveStyle.Lines || ly.Style == CurveStyle.LinesAndPoints || ly.Style == CurveStyle.Auto)
                                            gr.DrawLines(new Pen(ly.Color), pts);
                                        if (ly.Style == CurveStyle.Points || ly.Style == CurveStyle.LinesAndPoints || (ly.Style == CurveStyle.Auto && pts.Length <= 20))
                                        {
                                            Brush br = new SolidBrush(ly.Color);
                                            for (int j = 0; j < pts.Length; j++)
                                                gr.FillRectangle(br, pts[j].X - 2, pts[j].Y - 2, 5, 5);
                                        }
                                    }                                
                                    else if (layers[i] is HorizontalSelectionLayer)
                                    {
                                        HorizontalSelectionLayer ly = layers[i] as HorizontalSelectionLayer;
                                        float start = (float)ly.SelectionStart;
                                        float end = (float)ly.SelectionEnd;
                                        if (start < minX) start = minX;
                                        if (end > maxX) end = maxX;
                                        start = coordToCanvasX(start);
                                        end = coordToCanvasX(end);
                                        gr.FillRectangle(selectionBrush, start - 1, offsetY, end - start + 1, canvasHeight);
                                    }
                                    else if (layers[i] is HorizontalLineLayer)
                                    {
                                        throw new NotImplementedException();
                                    }
                                    else if (layers[i] is VerticalLineLayer)
                                    {
                                        throw new NotImplementedException();
                                    }
                                    else if (layers[i] is DataLabelLayer)
                                    {
                                        throw new NotImplementedException();
                                    }
                                    else if (layers[i] is WindowLabelLayer)
                                    {
                                        WindowLabelLayer ly = (layers[i] as WindowLabelLayer);
                                        float x = borderLeft;
                                        float y = borderTop;
                                        SizeF textSize = gr.MeasureString(ly.Text, this.Font);
                                        switch (ly.Alignment)
                                        {
                                            case ContentAlignment.TopLeft:
                                                x += textSize.Height;
                                                y += textSize.Height;
                                                break;
                                            case ContentAlignment.TopCenter:
                                                x += canvasWidth / 2 - textSize.Width / 2;
                                                y += textSize.Height;
                                                break;
                                            case ContentAlignment.TopRight:
                                                x += canvasWidth - textSize.Height - textSize.Width / 2;
                                                y += textSize.Height;
                                                break;
                                            default:
                                                throw new NotImplementedException();
                                        }
                                        gr.DrawString(ly.Text, this.Font, new SolidBrush(ly.Color), x, y + ly.Line * textSize.Height);
                                    }
                                    else if (layers[i] is CustomGraphicsLayer)
                                    {
                                        (layers[i] as CustomGraphicsLayer).CustomPaint(this, new PaintEventArgs(gr, new Rectangle(borderLeft, borderTop, canvasWidth, canvasHeight)));
                                    }
                                }
                            }
                        }
                    }

                    //Draw axes
                    if (needToDrawAxes)
                    {
                        gr.SetClip(pe.ClipRectangle);
                        int dpi2px = LogicalToDeviceUnits(2);
                        int dpi3px = LogicalToDeviceUnits(3);
                        int dpi4px = LogicalToDeviceUnits(4);

                        gr.DrawRectangle(axisPen, offsetX, offsetY, canvasWidth, canvasHeight);

                        //draw x axis label
                        string s;
                        float fht = gr.MeasureString("0", this.Font).Height;

                        if (logX)
                        {
                            f = (float)Math.Pow(10, Math.Floor(Math.Log10(minX)));
                            float fstep = f;
                            while (f < minX) f += fstep;
                            while (f <= maxX)
                            {
                                xc = coordToCanvasX(f);
                                f += fstep;
                                if (f >= 10 * fstep)
                                {
                                    f = (float)Math.Pow(10, Math.Floor(Math.Log10(f)));
                                    fstep = f;
                                    s = FormatFloatAxis(f, 0); //do not replace small numbers by 0
                                    gr.DrawString(s, this.Font, axisBrush, xc - gr.MeasureString(s, this.Font).Width / 2, this.Height - borderBottom + dpi3px);
                                }
                            }
                        }
                        else
                        {
                            for (int i = 0; i < axisTicksX.Length; i++)
                            {
                                f = axisTicksX[i];
                                xc = coordToCanvasX(f);
                                gr.DrawLine(axisPen, xc, this.Height - borderBottom - dpi4px, xc, this.Height - borderBottom);
                                gr.DrawLine(axisPen, xc, borderTop, xc, borderTop + dpi4px);
                                s = FormatFloatAxis(f, maxX - minX);
                                gr.DrawString(s, this.Font, axisBrush, xc - gr.MeasureString(s, this.Font).Width / 2, this.Height - borderBottom + dpi3px);
                            }
                        }
                        gr.DrawString(axisLabelX, this.Font, axisBrush, this.Width - borderRight - (canvasWidth + gr.MeasureString(axisLabelX, this.Font).Width) / 2, Height - fht - 1);

                        //draw y axis label
                        
                        if (logY)
                        {
                            f = (float)Math.Pow(10, Math.Floor(Math.Log10(minY)));
                            float fstep = f;
                            while (f < minY) f += fstep;
                            while (f <= maxY)
                            {
                                yc = coordToCanvasY(f);
                                gr.DrawLine(axisPen, borderLeft, yc, borderLeft + dpi2px, yc);
                                gr.DrawLine(axisPen, this.Width - borderRight - dpi2px, yc, this.Width - borderRight, yc);
                                f += fstep;
                                if (f >= 9.5f * fstep) fstep = f;
                            }
                        }
                        float span = (logY ? 0 : maxY - minY); //do not replace small numbers by 0
                        for (int i = 0; i < axisTicksY.Length; i++)
                        {
                            f = axisTicksY[i];
                            yc = coordToCanvasY(f);
                            gr.DrawLine(axisPen, borderLeft, yc, borderLeft + dpi4px, yc);
                            gr.DrawLine(axisPen, this.Width - borderRight - dpi4px, yc, this.Width - borderRight, yc);
                            s = FormatFloatAxis(f, span);
                            gr.DrawString(s, this.Font, axisBrush, borderLeft - dpi3px - gr.MeasureString(s, this.Font).Width, yc - fht / 2);
                        }

                        gr.RotateTransform(270);
                        gr.DrawString(axisLabelY, this.Font, axisBrush, -gr.MeasureString(axisLabelY, this.Font).Width / 2 - canvasHeight / 2 - borderTop, 0);
                        gr.ResetTransform();
                    }
                }
            }
            catch (Exception ex)
            {
                gr.Clip = oldclip;
                gr.DrawString("Cannot draw diagram: " + ex.Message, this.Font, Brushes.Red, borderLeft + 10, borderTop + 10);
            }
        }

        protected override void OnResize(EventArgs e)
        {
            base.OnResize(e);
            transformValid = false;
            this.Invalidate();
        }

        private string FormatFloatAxis(float f, float span)
        {
            if (Math.Abs(f * 1E6) < span) return "0"; //prevent label like 1.234E-15 instead of zero
            else if (Math.Abs(f) < 1E5 && Math.Abs(f) >= 1E-3) return f.ToString("0.#####");
            else return f.ToString("0.00E0");
        }

        protected override void ScaleControl(SizeF factor, BoundsSpecified specified)
        {
            SizeF currentScalingFactor = factor;
            base.ScaleControl(factor, specified);
            this.BorderBottom = (int)(this.BorderBottom * currentScalingFactor.Height + 0.5f);
            this.BorderTop = (int)(this.BorderTop * currentScalingFactor.Height + 0.5f);
            this.borderLeft = (int)(this.borderLeft * currentScalingFactor.Width + 0.5f);
            this.borderRight = (int)(this.borderRight * currentScalingFactor.Width + 0.5f);
        }
    }
}
