﻿using System;
using System.Collections.Generic;
using System.Runtime.InteropServices;
using Microsoft.Win32.SafeHandles;
using System.Security.Permissions;
using System.Runtime.ConstrainedExecution;

namespace DeviceClasses
{
    public class WinUsbInterface : CommunicationInterface, IUsbInterface
    {
		public static string[] FindInterfaces()
		{
            return FindInterfaces(new Guid("{a5dcbf10-6530-11d2-901f-00c04fb951ed}")); //GUID_DEVINTERFACE_USB_DEVICE
        }

        public static string[] FindInterfaces(Guid deviceIntefaceGuid)
        {
            List<string> devpaths = new List<string>();

			IntPtr deviceInfoSet = SetupDiGetClassDevs(ref deviceIntefaceGuid, IntPtr.Zero, IntPtr.Zero, DIGCF_PRESENT | DIGCF_DEVICEINTERFACE);
			try
			{
				SP_DEVICE_INTERFACE_DATA deviceInterfaceData = new SP_DEVICE_INTERFACE_DATA();
				deviceInterfaceData.cbSize = Marshal.SizeOf(deviceInterfaceData);
				Int32 bufferSize = 0;
				int i = 0;
				while (true)
				{
					bool success = SetupDiEnumDeviceInterfaces(deviceInfoSet, IntPtr.Zero, ref deviceIntefaceGuid, i, ref deviceInterfaceData);
					if (!success) break;
					SetupDiGetDeviceInterfaceDetail(deviceInfoSet, ref deviceInterfaceData, IntPtr.Zero, 0, ref bufferSize, IntPtr.Zero); //Note: this function never returns "success" here
					IntPtr detailDataBuffer = Marshal.AllocHGlobal(bufferSize);
					try
					{
						Marshal.WriteInt32(detailDataBuffer, (IntPtr.Size == 4) ? (4 + Marshal.SystemDefaultCharSize) : 8);
						success = SetupDiGetDeviceInterfaceDetail(deviceInfoSet, ref deviceInterfaceData, detailDataBuffer, bufferSize, ref bufferSize, IntPtr.Zero);
						if (!success) break;
						IntPtr pDevicePathName = new IntPtr(detailDataBuffer.ToInt64() + 4);
						devpaths.Add(Marshal.PtrToStringAuto(pDevicePathName));
					}
					finally
					{
						Marshal.FreeHGlobal(detailDataBuffer);
					}
					i++;
				}
			}
			finally
			{
				SetupDiDestroyDeviceInfoList(deviceInfoSet);
			}
			return devpaths.ToArray();
        }

        public WinUsbInterface(string devicePath)
        {
            this.devicePath = devicePath;

            deviceHandle = CreateFile(this.devicePath, GENERIC_WRITE | GENERIC_READ, FILE_SHARE_READ | FILE_SHARE_WRITE,
                IntPtr.Zero, OPEN_EXISTING, FILE_ATTRIBUTE_NORMAL | FILE_FLAG_OVERLAPPED, IntPtr.Zero);
			if (deviceHandle.IsInvalid)
			{
				isAvailable = false;
				System.Diagnostics.Debug.WriteLine("Cannot get handle to WinUSB interface (Win32Error:" + Marshal.GetLastWin32Error() + ").");
				return;
			}

			try
			{
				winUsbHandle = new SafeWinUsbHandle();
				bool success = WinUsb_Initialize(deviceHandle, ref winUsbHandle);
				if (!success) throw new Exception("Cannot initialize WinUSB interface.");

				try
				{
					uint transferred;
					uint size = (uint)Marshal.SizeOf(typeof(USB_DEVICE_DESCRIPTOR));
					success = WinUsb_GetDescriptor(winUsbHandle, USB_DEVICE_DESCRIPTOR_TYPE, 0, 0, out deviceDesc, size, out transferred);
					if (!success) throw new Exception("Failed to get USB device descriptor.");

					//Get supported language IDs
					byte[] buffer = new byte[256];
					success = WinUsb_GetDescriptor(winUsbHandle, USB_STRING_DESCRIPTOR_TYPE, 0, 0, buffer, (uint)buffer.Length, out transferred);
					if (!success) throw new CommunicationException("USB communication failure.");
					int length = buffer[0] - 2; //first byte is length, second byte is descriptor type
					langID = 0x0409; //default: US English
					if (length > 0) //Ocean Optics USB2000 workaround
					{
						ushort[] langIDs = new ushort[length / 2];
						Buffer.BlockCopy(buffer, 2, langIDs, 0, length);						
						if (langIDs.Length > 0) langID = langIDs[0]; // Get first supported language ID
					}

					manufacturer = GetStringDescriptor(deviceDesc.iManufacturer);
					product = GetStringDescriptor(deviceDesc.iProduct);
					serialNumber = GetStringDescriptor(deviceDesc.iSerialNumber);

                    //VID = deviceDesc.idVendor;
                    //PID = deviceDesc.idProduct;
                    //ClassValue = deviceDesc.bDeviceClass;
                    //SubClass = deviceDesc.bDeviceSubClass;
                    //Protocol = deviceDesc.bDeviceProtocol;

                    //Get maxPacketSize (not actually used here, but may be useful for other USB classes)
                    var pipeInfo = new WINUSB_PIPE_INFORMATION();
					success = WinUsb_QueryPipe(winUsbHandle, 0, 0, ref pipeInfo); //only check first pipe for maxPacketLength
					if (success) maxPacketSize = pipeInfo.MaximumPacketSize;
					else maxPacketSize = deviceDesc.bMaxPacketSize0;
				}
				finally
				{
					winUsbHandle.Close();
				}
			}
			finally
			{
				deviceHandle.Close();
			}
		}

		protected string GetStringDescriptor(byte index)
		{
			if (index == 0) return "?";
			byte[] buffer = new byte[256];
			uint transferred;
			bool success = WinUsb_GetDescriptor(winUsbHandle, USB_STRING_DESCRIPTOR_TYPE, index, langID, buffer, (uint)buffer.Length, out transferred);
			if (!success) throw new CommunicationException("Cannot get USB string descriptor (index " + index + ").");
			int length = buffer[0] - 2; //first byte is length, second byte is descriptor type
			return new string(System.Text.Encoding.Unicode.GetChars(buffer, 2, length));
		}

		public override void Open()
		{
			deviceHandle = CreateFile(devicePath, GENERIC_WRITE | GENERIC_READ, FILE_SHARE_READ | FILE_SHARE_WRITE,
						IntPtr.Zero, OPEN_EXISTING, FILE_ATTRIBUTE_NORMAL | FILE_FLAG_OVERLAPPED, IntPtr.Zero);
			if (deviceHandle.IsInvalid) throw new Exception("Cannot get handle to WinUSB interface.");

			winUsbHandle = new SafeWinUsbHandle();
			bool success = WinUsb_Initialize(deviceHandle, ref winUsbHandle);
			if (!success) throw new Exception("Cannot initialize WinUSB interface.");

			//Get total number of endpoints
			var ifaceDescriptors = new USB_INTERFACE_DESCRIPTOR[2];
			success = WinUsb_QueryInterfaceSettings(winUsbHandle, 0, ref ifaceDescriptors[0]);
			if (!success) throw new Exception("Cannot get WinUSB interface settings.");
			interfaceDesc = ifaceDescriptors[0];
			numEndpoints = interfaceDesc.bNumEndpoints;

			SetTimeout();

			if (numEndpoints > 0) //if there is any endpoint other than zero
			{
				uint val = Convert.ToUInt32(shortPacketTerminate);
				success = WinUsb_SetPipePolicy(winUsbHandle, outPipe, Convert.ToUInt32(POLICY_TYPE.SHORT_PACKET_TERMINATE), 4, ref val);
				if (!success) throw new Exception("Cannot set WinUSB pipe policy.");
			}

			//DisplayDeviceSpeed();
		}

		private void SetTimeout()
		{
			var pipeInfo = new WINUSB_PIPE_INFORMATION[2];
			for (int i = 0; i < numEndpoints; i++) //Set timeout for all endpoints
			{
				bool success = WinUsb_QueryPipe(winUsbHandle, 0, Convert.ToByte(i), ref pipeInfo[0]);
				if (!success) throw new Exception("Cannot query WinUSB pipe.");
				success = WinUsb_SetPipePolicy(winUsbHandle, pipeInfo[0].PipeId, Convert.ToUInt32(POLICY_TYPE.PIPE_TRANSFER_TIMEOUT), 4, ref pipeTimeout);
				if (!success) throw new Exception("Cannot set WinUSB pipe timeout.");
			}
		}

		public override void Close()
        {
			if (winUsbHandle != null) winUsbHandle.Close();
			if (deviceHandle != null) deviceHandle.Close();
		}

        public override bool IsOpen { get { return winUsbHandle != null && !winUsbHandle.IsInvalid; } }

		public bool IsAvailable { get { return isAvailable; } }
		protected bool isAvailable = true;

        public override string DevicePath { get { return devicePath; } }
        protected string devicePath;

		protected USB_DEVICE_DESCRIPTOR deviceDesc;
		protected USB_INTERFACE_DESCRIPTOR interfaceDesc;
		ushort langID;

		protected SafeFileHandle deviceHandle;
		protected SafeWinUsbHandle winUsbHandle;
		protected int numEndpoints;
		protected int maxPacketSize = 0;

		public uint Timeout
		{
			get { return pipeTimeout; }
			set
			{
				pipeTimeout = value;
				if (IsOpen) SetTimeout();
			}
		}
		protected uint pipeTimeout = 500;

		protected bool shortPacketTerminate = true;
		protected byte inPipe = 0x81;
		protected byte outPipe = 0x01;

		public string Product { get { return product; } }
		protected string product;

        public string Manufacturer { get { return manufacturer; } }
		protected string manufacturer;

        public string SerialNumber { get { return serialNumber; } }
		protected string serialNumber;

        public int VID { get { return (int)deviceDesc.idVendor; } }

		public int PID { get { return (int)deviceDesc.idProduct; } }

		public override void Write(byte[] buffer, int count)
        {
			UInt32 byteswritten;
            bool success = WinUsb_WritePipe(winUsbHandle, outPipe, buffer, (UInt32)count, out byteswritten, IntPtr.Zero);
			if (!success) throw new CommunicationException("USB write failed.");
		}

		public override int Read(byte[] buffer, int count)
        {
			UInt32 bytesread;
            bool success = WinUsb_ReadPipe(winUsbHandle, inPipe, buffer, (UInt32)count, out bytesread, IntPtr.Zero);
			if (!success) throw new CommunicationException("USB read failed.");
			return (int)bytesread;
		}

		public int ControlTransfer(byte RequestType, byte Request, ushort Value, ushort Index, ushort Length, ref byte[] Data) //Length may be 0
		{
			WINUSB_SETUP_PACKET setupPacket = new WINUSB_SETUP_PACKET() { RequestType = RequestType, Request = Request, Value = Value, Index = Index, Length = Length };
			uint bytesTransferred = 0;
			bool success = WinUsb_ControlTransfer(winUsbHandle, setupPacket, Data, Length, ref bytesTransferred, IntPtr.Zero);
			if (!success) throw new CommunicationException("Control transfer failed (Error " + System.Runtime.InteropServices.Marshal.GetLastWin32Error() + ").");
			return (int)bytesTransferred;
		}

        /*public static void FindDevicesWithWMI()
		{
			string scope = "root\\CIMV2"; //Root means root of the tree. CIMV2 means Common Information Model version 2.
			string query = "SELECT * FROM Win32_PnPEntity"; // The Win32_PnPEntity WMI class represents the properties of a Plug and Play device.
			ManagementObjectSearcher searcher = new ManagementObjectSearcher(scope, query);
			ManagementObjectCollection collection = searcher.Get();
			string[] WmiDeviceProperties = new string[] { "Name", "Caption", "Description", "Manufacturer", "PNPDeviceID", "DeviceID", "ClassGUID" };
			foreach (ManagementObject queryObj in collection)
			{
				if (queryObj["DeviceID"].ToString().StartsWith("USB\\"))
				{
					foreach (string devprop in WmiDeviceProperties) System.Diagnostics.Debug.WriteLine(devprop + ": " + queryObj[devprop]);
					System.Diagnostics.Debug.WriteLine("");
				}
			}
		}*/

		/// <summary>
		/// Handle to pass to WinUsb_Initialize.
		/// Adapted from the example in Microsoft's SafeHandle documentation.
		/// </summary>
		protected class SafeWinUsbHandle : SafeHandleZeroOrMinusOneIsInvalid
		{
			// Create a SafeHandle, informing the base class that this SafeHandle instance "owns" the handle,
			// and therefore SafeHandle should call our ReleaseHandle method when the SafeHandle is no longer in use.
			internal SafeWinUsbHandle() : base(true)
			{
				base.SetHandle(handle);
				this.handle = IntPtr.Zero;
			}

			/// <summary> Call WinUsb_Free on releasing the handle. </summary>
			/// <returns> True on success.</returns>
			protected override bool ReleaseHandle()
			{
				if (!this.IsInvalid) this.handle = IntPtr.Zero;
				return WinUsb_Free(handle);
			}

			public override bool IsInvalid
			{
				get { return (handle == IntPtr.Zero || handle == (IntPtr)(-1)); }
			}

			public IntPtr GetHandle()
			{
				if (IsInvalid) throw new Exception("The handle is invalid.");
				return handle;
			}
		}

        //************ Native Functions and Structures *************

            //WinUSB
        internal const UInt32 DEVICE_SPEED = 1;

		protected const int USB_DEVICE_DESCRIPTOR_TYPE = 0x01;
		protected const int USB_CONFIGURATION_DESCRIPTOR_TYPE = 0x02;
		protected const int USB_STRING_DESCRIPTOR_TYPE = 0x03;

		internal enum POLICY_TYPE
		{
			SHORT_PACKET_TERMINATE = 1,
			AUTO_CLEAR_STALL,
			PIPE_TRANSFER_TIMEOUT,
			IGNORE_SHORT_PACKETS,
			ALLOW_PARTIAL_READS,
			AUTO_FLUSH,
			RAW_IO,
		}

		internal enum USBD_PIPE_TYPE
		{
			UsbdPipeTypeControl,
			UsbdPipeTypeIsochronous,
			UsbdPipeTypeBulk,
			UsbdPipeTypeInterrupt,
		}

		internal enum USB_DEVICE_SPEED
		{
			UsbLowSpeed = 1,
			UsbFullSpeed,
			UsbHighSpeed,
		}

		[StructLayout(LayoutKind.Sequential)]
		protected struct USB_DEVICE_DESCRIPTOR
		{
			public byte bLength;
			public byte bDescriptorType;
			public ushort bcdUSB;
			public byte bDeviceClass;
			public byte bDeviceSubClass;
			public byte bDeviceProtocol;
			public byte bMaxPacketSize0;
			public ushort idVendor;
			public ushort idProduct;
			public ushort bcdDevice;
			public byte iManufacturer;
			public byte iProduct;
			public byte iSerialNumber;
			public byte bNumConfigurations;
		}

		[StructLayout(LayoutKind.Sequential)]
		protected struct USB_CONFIGURATION_DESCRIPTOR
		{
			internal Byte bLength;
			internal Byte bDescriptorType;
			internal UInt16 wTotalLength;
			internal Byte bNumInterfaces;
			internal Byte bConfigurationValue;
			internal Byte iConfiguration;
			internal Byte bmAttributes;
			internal Byte MaxPower;
		}

		[StructLayout(LayoutKind.Sequential)]
		protected struct USB_INTERFACE_DESCRIPTOR
		{
			internal Byte bLength;
			internal Byte bDescriptorType;
			internal Byte bInterfaceNumber;
			internal Byte bAlternateSetting;
			internal Byte bNumEndpoints;
			internal Byte bInterfaceClass;
			internal Byte bInterfaceSubClass;
			internal Byte bInterfaceProtocol;
			internal Byte iInterface;
		}

		[StructLayout(LayoutKind.Sequential)]
		protected struct WINUSB_PIPE_INFORMATION
		{
			internal USBD_PIPE_TYPE PipeType;
			internal Byte PipeId;
			internal UInt16 MaximumPacketSize;
			internal Byte Interval;
		}

		[StructLayout(LayoutKind.Sequential)]
		protected struct USBD_ISO_PACKET_DESCRIPTOR
		{
			internal UInt32 Offset;
			internal UInt32 Length;
			internal UInt32 Status;
		}

		[StructLayout(LayoutKind.Sequential, Pack = 1)]
		protected struct WINUSB_SETUP_PACKET
		{
			internal Byte RequestType;
			internal Byte Request;
			internal UInt16 Value;
			internal UInt16 Index;
			internal UInt16 Length;
		}

		[DllImport("winusb.dll", SetLastError = true)]
		protected static extern Boolean WinUsb_ControlTransfer(SafeWinUsbHandle InterfaceHandle, WINUSB_SETUP_PACKET SetupPacket, Byte[] Buffer, UInt32 BufferLength, ref UInt32 LengthTransferred, IntPtr Overlapped);

		[DllImport("winusb.dll", SetLastError = true)]
		protected static extern Boolean WinUsb_Free(IntPtr InterfaceHandle);

		[DllImport("winusb.dll", SetLastError = true)]
		protected static extern Boolean WinUsb_Initialize(SafeFileHandle DeviceHandle, ref SafeWinUsbHandle InterfaceHandle);

		[DllImport("winusb.dll", SetLastError = true)]
		protected static extern Boolean WinUsb_QueryDeviceInformation(SafeWinUsbHandle InterfaceHandle, UInt32 InformationType, ref UInt32 BufferLength, ref Byte Buffer);

		[DllImport("winusb.dll", SetLastError = true)]
		protected static extern Boolean WinUsb_QueryInterfaceSettings(SafeWinUsbHandle InterfaceHandle, Byte AlternateInterfaceNumber, ref USB_INTERFACE_DESCRIPTOR UsbAltInterfaceDescriptor);

		[DllImport("winusb.dll", SetLastError = true)]
		protected static extern Boolean WinUsb_QueryPipe(SafeWinUsbHandle InterfaceHandle, Byte AlternateInterfaceNumber, Byte PipeIndex, ref WINUSB_PIPE_INFORMATION PipeInformation);

		[DllImport("winusb.dll", SetLastError = true)]
		protected static extern Boolean WinUsb_ReadPipe(SafeWinUsbHandle InterfaceHandle, Byte PipeID, Byte[] Buffer, UInt32 BufferLength, out UInt32 LengthTransferred, IntPtr Overlapped);

		[DllImport("winusb.dll", SetLastError = true)]
		protected static extern Boolean WinUsb_SetPipePolicy(SafeWinUsbHandle InterfaceHandle, Byte PipeID, UInt32 PolicyType, UInt32 ValueLength, ref UInt32 Value);

		[DllImport("winusb.dll", SetLastError = true)]
		protected static extern Boolean WinUsb_WritePipe(SafeWinUsbHandle InterfaceHandle, Byte PipeID, Byte[] Buffer, UInt32 BufferLength, out UInt32 LengthTransferred, IntPtr Overlapped);

		[DllImport("winusb.dll", SetLastError = true)]
		protected static extern bool WinUsb_GetDescriptor(SafeWinUsbHandle InterfaceHandle, byte DescriptorType, byte Index, UInt16 LanguageID, byte[] Buffer, UInt32 BufferLength, out UInt32 LengthTransfered);

		[DllImport("winusb.dll", SetLastError = true)]
		private static extern bool WinUsb_GetDescriptor(SafeWinUsbHandle InterfaceHandle, byte DescriptorType, byte Index, UInt16 LanguageID, out USB_DEVICE_DESCRIPTOR deviceDesc, UInt32 BufferLength, out UInt32 LengthTransfered);

		[DllImport("winusb.dll", SetLastError = true)]
		private static extern bool WinUsb_GetDescriptor(SafeWinUsbHandle InterfaceHandle, byte DescriptorType, byte Index, UInt16 LanguageID, out USB_CONFIGURATION_DESCRIPTOR deviceDesc, UInt32 BufferLength, out UInt32 LengthTransfered);

		//SetupApi

		internal const Int32 DIGCF_PRESENT = 2;
		internal const Int32 DIGCF_DEVICEINTERFACE = 0X10;

		internal struct SP_DEVICE_INTERFACE_DATA
		{
			internal Int32 cbSize;
			internal Guid InterfaceClassGuid;
			internal Int32 Flags;
			internal IntPtr Reserved;
		}

		[DllImport("setupapi.dll", SetLastError = true, CharSet = CharSet.Auto)]
		internal static extern IntPtr SetupDiGetClassDevs(ref Guid ClassGuid, IntPtr Enumerator, IntPtr hwndParent, Int32 Flags);

		[DllImport("setupapi.dll", SetLastError = true)]
		internal static extern Boolean SetupDiEnumDeviceInterfaces(IntPtr DeviceInfoSet, IntPtr DeviceInfoData, ref Guid InterfaceClassGuid, Int32 MemberIndex, ref SP_DEVICE_INTERFACE_DATA DeviceInterfaceData);

		[DllImport("setupapi.dll", SetLastError = true, CharSet = CharSet.Auto)]
		internal static extern Boolean SetupDiGetDeviceInterfaceDetail(IntPtr DeviceInfoSet, ref SP_DEVICE_INTERFACE_DATA DeviceInterfaceData, IntPtr DeviceInterfaceDetailData, Int32 DeviceInterfaceDetailDataSize, ref Int32 RequiredSize, IntPtr DeviceInfoData);

		[DllImport("setupapi.dll", SetLastError = true)]
		internal static extern Int32 SetupDiDestroyDeviceInfoList(IntPtr DeviceInfoSet);

		//File I/O

		internal const Int32 FILE_ATTRIBUTE_NORMAL = 0X80;
		internal const Int32 FILE_FLAG_OVERLAPPED = 0X40000000;
		internal const Int32 FILE_SHARE_READ = 1;
		internal const Int32 FILE_SHARE_WRITE = 2;
		internal const UInt32 GENERIC_READ = 0X80000000;
		internal const UInt32 GENERIC_WRITE = 0X40000000;
		internal const Int32 INVALID_HANDLE_VALUE = -1;
		internal const Int32 OPEN_EXISTING = 3;

		[DllImport("kernel32.dll", CharSet = CharSet.Auto, SetLastError = true)]
		internal static extern SafeFileHandle CreateFile(String lpFileName, UInt32 dwDesiredAccess, Int32 dwShareMode, IntPtr lpSecurityAttributes, Int32 dwCreationDisposition, Int32 dwFlagsAndAttributes, IntPtr hTemplateFile);
    }
}
