﻿using AuxControls;
using DeviceClasses;
using System;
using System.Drawing;
using System.IO;
using System.Windows.Forms;
using System.Globalization;
using System.Threading.Tasks;
using System.Diagnostics.Eventing.Reader;

namespace SimpleSpectrometer
{
    public partial class MainForm : Form
    {
        private ZioLinkSpectrometer spectr;
        private SpectrumData spectrum;
        private DateTime spectrumDate;
        bool updatingUI = false;

        public MainForm()
        {
            InitializeComponent();
            UpdateStatusInUI(SpectrStatus.NotConnected);
            deviceNameLabel.Text = "";
            deviceSerialNoLabel.Text = "";
        }

        private void openButton_Click(object sender, EventArgs e)
        {
            ZioLinkSpectrometer[] devices = ZioLinkSpectrometer.FindDevices();
            if (devices.Length == 0)
                deviceLabel.Text = "No device found.";
            else
            {
                spectr = devices[0]; //only consider first device

                updatingUI = true;
                try
                {
                    spectr.Open();

                    spectr.AutoExposure = false;
                    expTimeBox.Maximum = (decimal)(spectr.MaxExposureTime * 1000f);
                    expTimeBox.Minimum = (decimal)(spectr.MinExposureTime * 1000f);
                    expTimeBox.Value = (decimal)(spectr.ExposureTime * 1000f);
                    averagingBox.Maximum = (decimal)spectr.MaxAveraging;
                    averagingBox.Value = spectr.Averaging;
                    UpdateStatusInUI(spectr.GetStatus());

                    deviceLabel.Text = "Device found:";
                    deviceNameLabel.Text = spectr.DeviceName;
                    deviceSerialNoLabel.Text = "s/n: " + spectr.SerialNumber;
                    openButton.Enabled = false;
                    expTimeBox.Enabled = true;
                    averagingBox.Enabled = true;
                    takeSpectrumButton.Enabled = true;
                    triggerDelayBox.Enabled = true;

                    //Initialize trigger user interface
                    if (spectr.TriggerMode == ZioLinkSpectrometer.TriggerModes.HardwareTrigger) hardwareTriggerRadioButton.Checked = true;
                    else freeRunningRadioButton.Checked = true;
                    externalTriggerCheckBox.Checked = spectr.UseExternalTrigger;
                    edgeSelectionComboBox.SelectedIndex = (int)spectr.TriggerEdge;
                    triggerDelayBox.Maximum = (decimal)spectr.MaxTriggerDelay * 1000; //convert from s to ms
                    triggerDelayBox.Minimum = (decimal)spectr.MinTriggerDelay * 1000; //convert from s to ms
                    triggerDelayBox.Value = (decimal)spectr.TriggerDelay * 1000; //convert from s to ms
                }
                catch (Exception ex)
                {
                    deviceLabel.Text = ex.Message;
                    MessageBox.Show(ex.Message, "Cannot open spectrometer", MessageBoxButtons.OK, MessageBoxIcon.Error);
                }
                updatingUI = false;
            }
        }

        private void expTimeBox_ValueChanged(object sender, EventArgs e)
        {
            if (updatingUI) return; //if value is changed programatically, not by user action
            try
            {
                spectr.ExposureTime = (float)expTimeBox.Value * 0.001f;
            }
            catch (Exception ex)
            {
                MessageBox.Show(ex.Message, "Cannot set exposure time", MessageBoxButtons.OK, MessageBoxIcon.Error);
            }
        }

        private void averagingBox_ValueChanged(object sender, EventArgs e)
        {
            if (updatingUI) return; //if value is changed programatically, not by user action
            try
            {
                spectr.Averaging = (int)averagingBox.Value;
            }
            catch (Exception ex)
            {
                MessageBox.Show(ex.Message, "Cannot set averaging", MessageBoxButtons.OK, MessageBoxIcon.Error);
            }
        }

        private void triggerModeRadioButton_Click(object sender, EventArgs e)
        {
            try
            {
                (spectr as ZioLinkSpectrometer).TriggerMode = (freeRunningRadioButton.Checked ? ZioLinkSpectrometer.TriggerModes.FreeRunningTriggerEnd : ZioLinkSpectrometer.TriggerModes.HardwareTrigger);
            }
            catch (Exception ex)
            {
                MessageBox.Show(ex.Message, "Cannot configure trigger", MessageBoxButtons.OK, MessageBoxIcon.Error);
            }
        }

        private void externalTriggerCheckBox_Click(object sender, EventArgs e)
        {
            try
            {
                (spectr as ZioLinkSpectrometer).UseExternalTrigger = externalTriggerCheckBox.Checked;
            }
            catch (Exception ex)
            {
                MessageBox.Show(ex.Message, "Cannot set external trigger", MessageBoxButtons.OK, MessageBoxIcon.Error);
            }
        }

        private void edgeSelectionComboBox_SelectionChangeCommitted(object sender, EventArgs e)
        {
            try
            {
                (spectr as ZioLinkSpectrometer).TriggerEdge = (ZioLinkSpectrometer.TriggerEdges)edgeSelectionComboBox.SelectedIndex;
            }
            catch (Exception ex)
            {
                MessageBox.Show(ex.Message, "Cannot configure trigger", MessageBoxButtons.OK, MessageBoxIcon.Error);
            }
        }

        private void triggerDelayFloatBox_ValueChanged(object sender, EventArgs e)
        {
            if (updatingUI) return; //if value is changed programatically, not by user action
            try
            {
                (spectr as ZioLinkSpectrometer).TriggerDelay = (float)triggerDelayBox.Value * 1E-3f; //convert from ms to s
            }
            catch (Exception ex)
            {
                MessageBox.Show(ex.Message, "Cannot set trigger in delay", MessageBoxButtons.OK, MessageBoxIcon.Error);
            }
        }

        private void takeSpectrumButton_Click(object sender, EventArgs e)
        {
            try
            {
                if (readoutTimer.Enabled) //if exposure is in progress
                {
                    spectr.CancelExposure();
                    readoutTimer.Enabled = false;
                    UpdateStatusInUI(spectr.GetStatus());
                }
                else
                {
                    spectr.StartExposure(1); //take one spectrum
                    readoutTimer.Enabled = true;
                }
            }
            catch (Exception ex)
            {
                MessageBox.Show(ex.Message, "Cannot start/stop exposure", MessageBoxButtons.OK, MessageBoxIcon.Error);
            }
        }

        private void readoutTimer_Tick(object sender, EventArgs e)
        {
            try
            {
                SpectrStatus st = spectr.GetStatus(out int available);
                UpdateStatusInUI(st);

                for (int i = 0; i < available; i++) //Loop not really required in this case
                {
                    spectrum = spectr.GetSpectrumData();
                    spectrumDate = DateTime.Now;
                    displaySpectrum();
                    saveButton.Enabled = true;
                }
                if (st == SpectrStatus.Idle) readoutTimer.Enabled = false;
            }
            catch (Exception ex)
            {
                readoutTimer.Enabled = false;
                MessageBox.Show(ex.Message, "Cannot read spectrum from device", MessageBoxButtons.OK, MessageBoxIcon.Error);
            }
        }

        private void UpdateStatusInUI(SpectrStatus Status)
        {
            if (Status == SpectrStatus.Idle) takeSpectrumButton.Text = "Take Spectrum";
            else takeSpectrumButton.Text = "Stop";

            statusLabel.Text = statusStr[(int)Status];

            bool isOpen = (spectr?.IsOpen == true);
            expTimeBox.Enabled = isOpen;
            averagingBox.Enabled = isOpen;

            bool canChangeTrigger = isOpen && (spectr is ZioLinkSpectrometer) && (Status == SpectrStatus.Idle);
            freeRunningRadioButton.Enabled = canChangeTrigger;
            hardwareTriggerRadioButton.Enabled = canChangeTrigger;
            externalTriggerCheckBox.Enabled = canChangeTrigger;
            edgeSelectionComboBox.Enabled = canChangeTrigger;
        }

        private string[] statusStr = new string[] { "Idle", "Waiting for trigger", "Taking spectrum",
            "Waiting for temperature", "?", "?", "?", "?", "PoweredOff", "SleepMode", "" };

        private void displaySpectrum()
        {
            int pixelCount = spectrum.PixelCount;

            spectrumDiagram.Layers.Clear();
            spectrumDiagram.Layers.Add(new Diagram.SinglePrecDataLayer(
                spectr.Wavelengths, spectrum.Values, Color.DarkRed, Diagram.CurveStyle.Lines));

            spectrumDiagram.MinX = spectr.Wavelengths[0];
            spectrumDiagram.MaxX = spectr.Wavelengths[pixelCount - 1];

            float min = float.MaxValue;
            float max = float.MinValue;
            for (int i = 0; i < pixelCount; i++)
            {
                if (spectrum.Values[i] < min) min = spectrum.Values[i];
                if (spectrum.Values[i] > max) max = spectrum.Values[i];
            }
            spectrumDiagram.MinY = Math.Min(min, 0);
            spectrumDiagram.MaxY = max * 1.04f;

            spectrumDiagram.Invalidate(); //repaint diagram
        }

        private void saveButton_Click(object sender, EventArgs e)
        {
            using (SaveFileDialog dlg = new SaveFileDialog())
            {
                dlg.Title = "Save Spectrum File";
                dlg.Filter = "Spectrum files(*.spz, *.spc)|*.spz;*.spc|All files|*.*";
                dlg.DefaultExt = "spz";

                if (dlg.ShowDialog() == DialogResult.OK)
                {
                    Cursor.Current = Cursors.WaitCursor;
                    try
                    {
                        using (FileStream fs = new FileStream(dlg.FileName, FileMode.Create, FileAccess.Write))
                        {
                            using (StreamWriter sw = new StreamWriter(fs))
                            {
                                sw.WriteLine("SPECTRUM DATA FILE");
                                sw.WriteLine("SpectrumCount:\t1");
                                sw.WriteLine("PixelCount:\t" + spectrum.PixelCount);
                                sw.WriteLine("WavelengthUnit:\tnm");
                                sw.WriteLine("DeviceName:\t" + spectr.DeviceName);
                                sw.WriteLine("DeviceManufacturer:\t" + spectr.Manufacturer);
                                sw.WriteLine("DeviceSerialNo:\t" + spectr.SerialNumber);
                                sw.WriteLine();

                                sw.WriteLine("Name:\tSpectrum");
                                sw.WriteLine("Date:\t" + spectrumDate);
                                sw.WriteLine("ExposureTime:\t" + spectrum.ExposureTime.ToString(CultureInfo.InvariantCulture));
                                sw.WriteLine("Averaging:\t" + spectrum.Averaging.ToString());
                                sw.WriteLine("LoadLevel:\t" + spectrum.LoadLevel.ToString(CultureInfo.InvariantCulture));
                                sw.WriteLine("Temperature:\t" + spectrum.SensorTemperature.ToString(CultureInfo.InvariantCulture));
                                sw.WriteLine("ExposureSettings:\t" + spectrum.ExposureSettings);
                                sw.WriteLine("AppliedProcessing:\t" + (int)spectrum.AppliedProcessing);
                                sw.WriteLine();
                                sw.WriteLine("Wavelength\tSpectrum");
                                for (int p = 0; p < spectrum.PixelCount; p++)
                                {
                                    sw.Write((spectr.Wavelengths[p]).ToString(CultureInfo.InvariantCulture));
                                    sw.Write('\t');
                                    sw.Write(((float)spectrum.Values[p]).ToString(CultureInfo.InvariantCulture));
                                    sw.WriteLine();
                                }
                            }
                        }
                    }
                    catch (Exception ex)
                    {
                        MessageBox.Show(ex.Message, "Cannot Save File", MessageBoxButtons.OK, MessageBoxIcon.Error);
                    }
                    Cursor.Current = Cursors.Default;
                }
            }
        }

        private void closeButton_Click(object sender, EventArgs e)
        {
            if (spectr != null) spectr.Close(); //does not throw exceptions
            Close();
        }
    }
}
